cimport cython
from dataclasses import dataclass, field
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Union,
)

from apischema.aliases import Aliaser
from apischema.conversions.utils import Converter
from apischema.deserialization.coercion import Coercer
from apischema.json_schema.types import bad_type
from apischema.types import AnyType, NoneType
from apischema.utils import Lazy
from apischema.validation.errors import (
    ErrorKey,
    ErrorMsg,
    ValidationError,
    merge_errors,
)
from apischema.validation.mock import ValidatorMock
from apischema.validation.validators import Validator, validate


cdef class Constraint():
    cdef readonly object error
    cdef int _dispatch

    def __init__(self, error):
        self.error = error

cdef class MinimumConstraint(Constraint):
    cdef readonly long minimum

    cpdef validate(self, object data):
        return MinimumConstraint_validate(self, data)

    def __init__(self, error, minimum):
        self.error = error
        self.minimum = minimum
        self._dispatch = 0

cdef class MaximumConstraint(Constraint):
    cdef readonly long maximum

    cpdef validate(self, object data):
        return MaximumConstraint_validate(self, data)

    def __init__(self, error, maximum):
        self.error = error
        self.maximum = maximum
        self._dispatch = 1

cdef class ExclusiveMinimumConstraint(Constraint):
    cdef readonly long exc_min

    cpdef validate(self, object data):
        return ExclusiveMinimumConstraint_validate(self, data)

    def __init__(self, error, exc_min):
        self.error = error
        self.exc_min = exc_min
        self._dispatch = 2

cdef class ExclusiveMaximumConstraint(Constraint):
    cdef readonly long exc_max

    cpdef validate(self, object data):
        return ExclusiveMaximumConstraint_validate(self, data)

    def __init__(self, error, exc_max):
        self.error = error
        self.exc_max = exc_max
        self._dispatch = 3

cdef class MultipleOfConstraint(Constraint):
    cdef readonly long mult_of

    cpdef validate(self, object data):
        return MultipleOfConstraint_validate(self, data)

    def __init__(self, error, mult_of):
        self.error = error
        self.mult_of = mult_of
        self._dispatch = 4

cdef class MinLengthConstraint(Constraint):
    cdef readonly long min_len

    cpdef validate(self, object data):
        return MinLengthConstraint_validate(self, data)

    def __init__(self, error, min_len):
        self.error = error
        self.min_len = min_len
        self._dispatch = 5

cdef class MaxLengthConstraint(Constraint):
    cdef readonly long max_len

    cpdef validate(self, object data):
        return MaxLengthConstraint_validate(self, data)

    def __init__(self, error, max_len):
        self.error = error
        self.max_len = max_len
        self._dispatch = 6

cdef class PatternConstraint(Constraint):
    cdef readonly object pattern

    cpdef validate(self, object data):
        return PatternConstraint_validate(self, data)

    def __init__(self, error, pattern):
        self.error = error
        self.pattern = pattern
        self._dispatch = 7

cdef class MinItemsConstraint(Constraint):
    cdef readonly long min_items

    cpdef validate(self, object data):
        return MinItemsConstraint_validate(self, data)

    def __init__(self, error, min_items):
        self.error = error
        self.min_items = min_items
        self._dispatch = 8

cdef class MaxItemsConstraint(Constraint):
    cdef readonly long max_items

    cpdef validate(self, object data):
        return MaxItemsConstraint_validate(self, data)

    def __init__(self, error, max_items):
        self.error = error
        self.max_items = max_items
        self._dispatch = 9

cdef class UniqueItemsConstraint(Constraint):
    cdef readonly bint unique

    cpdef validate(self, object data):
        return UniqueItemsConstraint_validate(self, data)

    def __init__(self, error, unique):
        self.error = error
        self.unique = unique
        assert self.unique
        self._dispatch = 10

cdef class MinPropertiesConstraint(Constraint):
    cdef readonly long min_properties

    cpdef validate(self, object data):
        return MinPropertiesConstraint_validate(self, data)

    def __init__(self, error, min_properties):
        self.error = error
        self.min_properties = min_properties
        self._dispatch = 11

cdef class MaxPropertiesConstraint(Constraint):
    cdef readonly long max_properties

    cpdef validate(self, object data):
        return MaxPropertiesConstraint_validate(self, data)

    def __init__(self, error, max_properties):
        self.error = error
        self.max_properties = max_properties
        self._dispatch = 12

cdef class DeserializationMethod():
    cdef int _dispatch

cdef class RecMethod(DeserializationMethod):
    cdef readonly object lazy
    cdef readonly object method

    cpdef deserialize(self, object data):
        return RecMethod_deserialize(self, data)

    def __init__(self, lazy):
        self.lazy = lazy
        self.method = None
        self._dispatch = 0

cdef class ValidatorMethod(DeserializationMethod):
    cdef readonly DeserializationMethod method
    cdef readonly object validators
    cdef readonly object aliaser

    cpdef deserialize(self, object data):
        return ValidatorMethod_deserialize(self, data)

    def __init__(self, method, validators, aliaser):
        self.method = method
        self.validators = validators
        self.aliaser = aliaser
        self._dispatch = 1

cdef class CoercerMethod(DeserializationMethod):
    cdef readonly object coercer
    cdef readonly type cls
    cdef readonly DeserializationMethod method

    cpdef deserialize(self, object data):
        return CoercerMethod_deserialize(self, data)

    def __init__(self, coercer, cls, method):
        self.coercer = coercer
        self.cls = cls
        self.method = method
        self._dispatch = 2

cdef class TypeCheckMethod(DeserializationMethod):
    cdef readonly object expected
    cdef readonly DeserializationMethod fallback

    cpdef deserialize(self, object data):
        return TypeCheckMethod_deserialize(self, data)

    def __init__(self, expected, fallback):
        self.expected = expected
        self.fallback = fallback
        self._dispatch = 3

cdef class AnyMethod(DeserializationMethod):
    cdef readonly dict constraints

    cpdef deserialize(self, object data):
        return AnyMethod_deserialize(self, data)

    def __init__(self, constraints):
        self.constraints = constraints
        self._dispatch = 4

cdef class ListCheckOnlyMethod(DeserializationMethod):
    cdef readonly tuple constraints
    cdef readonly DeserializationMethod value_method

    cpdef deserialize(self, object data):
        return ListCheckOnlyMethod_deserialize(self, data)

    def __init__(self, constraints, value_method):
        self.constraints = constraints
        self.value_method = value_method
        self._dispatch = 5

cdef class ListMethod(DeserializationMethod):
    cdef readonly tuple constraints
    cdef readonly DeserializationMethod value_method

    cpdef deserialize(self, object data):
        return ListMethod_deserialize(self, data)

    def __init__(self, constraints, value_method):
        self.constraints = constraints
        self.value_method = value_method
        self._dispatch = 6

cdef class SetMethod(DeserializationMethod):
    cdef readonly tuple constraints
    cdef readonly DeserializationMethod value_method

    cpdef deserialize(self, object data):
        return SetMethod_deserialize(self, data)

    def __init__(self, constraints, value_method):
        self.constraints = constraints
        self.value_method = value_method
        self._dispatch = 7

cdef class VariadicTupleMethod(DeserializationMethod):
    cdef readonly DeserializationMethod method

    cpdef deserialize(self, object data):
        return VariadicTupleMethod_deserialize(self, data)

    def __init__(self, method):
        self.method = method
        self._dispatch = 8

cdef class LiteralMethod(DeserializationMethod):
    cdef readonly dict value_map
    cdef readonly object error
    cdef readonly object coercer
    cdef readonly tuple types

    cpdef deserialize(self, object data):
        return LiteralMethod_deserialize(self, data)

    def __init__(self, value_map, error, coercer, types):
        self.value_map = value_map
        self.error = error
        self.coercer = coercer
        self.types = types
        self._dispatch = 9

cdef class MappingCheckOnly(DeserializationMethod):
    cdef readonly tuple constraints
    cdef readonly DeserializationMethod key_method
    cdef readonly DeserializationMethod value_method

    cpdef deserialize(self, object data):
        return MappingCheckOnly_deserialize(self, data)

    def __init__(self, constraints, key_method, value_method):
        self.constraints = constraints
        self.key_method = key_method
        self.value_method = value_method
        self._dispatch = 10

cdef class MappingMethod(DeserializationMethod):
    cdef readonly tuple constraints
    cdef readonly DeserializationMethod key_method
    cdef readonly DeserializationMethod value_method

    cpdef deserialize(self, object data):
        return MappingMethod_deserialize(self, data)

    def __init__(self, constraints, key_method, value_method):
        self.constraints = constraints
        self.key_method = key_method
        self.value_method = value_method
        self._dispatch = 11

cdef class Field():
    cdef readonly str name
    cdef readonly str alias
    cdef readonly DeserializationMethod method
    cdef readonly bint required
    cdef readonly object required_by
    cdef readonly bint fall_back_on_default

    def __init__(self, name, alias, method, required, required_by, fall_back_on_default):
        self.name = name
        self.alias = alias
        self.method = method
        self.required = required
        self.required_by = required_by
        self.fall_back_on_default = fall_back_on_default

cdef class FlattenedField():
    cdef readonly str name
    cdef readonly tuple aliases
    cdef readonly DeserializationMethod method
    cdef readonly bint fall_back_on_default

    def __init__(self, name, aliases, method, fall_back_on_default):
        self.name = name
        self.aliases = aliases
        self.method = method
        self.fall_back_on_default = fall_back_on_default

cdef class PatternField():
    cdef readonly str name
    cdef readonly object pattern
    cdef readonly DeserializationMethod method
    cdef readonly bint fall_back_on_default

    def __init__(self, name, pattern, method, fall_back_on_default):
        self.name = name
        self.pattern = pattern
        self.method = method
        self.fall_back_on_default = fall_back_on_default

cdef class AdditionalField():
    cdef readonly str name
    cdef readonly DeserializationMethod method
    cdef readonly bint fall_back_on_default

    def __init__(self, name, method, fall_back_on_default):
        self.name = name
        self.method = method
        self.fall_back_on_default = fall_back_on_default

cdef class Constructor():
    cdef readonly object cls
    cdef int _dispatch

    def __init__(self, cls):
        self.cls = cls

cdef class NoConstructor(Constructor):

    cpdef construct(self, dict fields):
        return NoConstructor_construct(self, fields)

    def __init__(self, cls):
        self.cls = cls
        self._dispatch = 0

cdef class RawConstructor(Constructor):

    cpdef construct(self, dict fields):
        return RawConstructor_construct(self, fields)

    def __init__(self, cls):
        self.cls = cls
        self._dispatch = 1

cdef class DefaultField():
    cdef readonly str name
    cdef readonly object default_value

    def __init__(self, name, default_value):
        self.name = name
        self.default_value = default_value

cdef class FactoryField():
    cdef readonly str name
    cdef readonly object factory

    def __init__(self, name, factory):
        self.name = name
        self.factory = factory

cdef class FieldsConstructor(Constructor):
    cdef readonly long nb_fields
    cdef readonly tuple default_fields
    cdef readonly tuple factory_fields

    cpdef construct(self, dict fields):
        return FieldsConstructor_construct(self, fields)

    def __init__(self, cls, nb_fields, default_fields, factory_fields):
        self.cls = cls
        self.nb_fields = nb_fields
        self.default_fields = default_fields
        self.factory_fields = factory_fields
        self._dispatch = 2

cdef class SimpleObjectMethod(DeserializationMethod):
    cdef readonly Constructor constructor
    cdef readonly tuple fields
    cdef readonly set all_aliases
    cdef readonly bint typed_dict
    cdef readonly str missing
    cdef readonly str unexpected

    cpdef deserialize(self, object data):
        return SimpleObjectMethod_deserialize(self, data)

    def __init__(self, constructor, fields, all_aliases, typed_dict, missing, unexpected):
        self.constructor = constructor
        self.fields = fields
        self.all_aliases = all_aliases
        self.typed_dict = typed_dict
        self.missing = missing
        self.unexpected = unexpected
        self._dispatch = 12

cdef class ObjectMethod(DeserializationMethod):
    cdef readonly Constructor constructor
    cdef readonly tuple constraints
    cdef readonly tuple fields
    cdef readonly tuple flattened_fields
    cdef readonly tuple pattern_fields
    cdef readonly object additional_field
    cdef readonly set all_aliases
    cdef readonly bint additional_properties
    cdef readonly bint typed_dict
    cdef readonly tuple validators
    cdef readonly tuple init_defaults
    cdef readonly set post_init_modified
    cdef readonly object aliaser
    cdef readonly str missing
    cdef readonly str unexpected
    cdef readonly object discriminator
    cdef readonly bint aggregate_fields

    cpdef deserialize(self, object data):
        return ObjectMethod_deserialize(self, data)

    def __init__(self, constructor, constraints, fields, flattened_fields, pattern_fields, additional_field, all_aliases, additional_properties, typed_dict, validators, init_defaults, post_init_modified, aliaser, missing, unexpected, discriminator):
        self.constructor = constructor
        self.constraints = constraints
        self.fields = fields
        self.flattened_fields = flattened_fields
        self.pattern_fields = pattern_fields
        self.additional_field = additional_field
        self.all_aliases = all_aliases
        self.additional_properties = additional_properties
        self.typed_dict = typed_dict
        self.validators = validators
        self.init_defaults = init_defaults
        self.post_init_modified = post_init_modified
        self.aliaser = aliaser
        self.missing = missing
        self.unexpected = unexpected
        self.discriminator = discriminator
        self.aggregate_fields = bool(
            self.flattened_fields
            or self.pattern_fields
            or self.additional_field is not None
        )
        self._dispatch = 13

cdef class NoneMethod(DeserializationMethod):

    cpdef deserialize(self, object data):
        return NoneMethod_deserialize(self, data)

    def __init__(self):
        self._dispatch = 14

cdef class IntMethod(DeserializationMethod):

    cpdef deserialize(self, object data):
        return IntMethod_deserialize(self, data)

    def __init__(self):
        self._dispatch = 15

cdef class FloatMethod(DeserializationMethod):

    cpdef deserialize(self, object data):
        return FloatMethod_deserialize(self, data)

    def __init__(self):
        self._dispatch = 17

cdef class StrMethod(DeserializationMethod):

    cpdef deserialize(self, object data):
        return StrMethod_deserialize(self, data)

    def __init__(self):
        self._dispatch = 19

cdef class BoolMethod(DeserializationMethod):

    cpdef deserialize(self, object data):
        return BoolMethod_deserialize(self, data)

    def __init__(self):
        self._dispatch = 21

cdef class ConstrainedIntMethod(IntMethod):
    cdef readonly tuple constraints

    cpdef deserialize(self, object data):
        return ConstrainedIntMethod_deserialize(self, data)

    def __init__(self, constraints):
        self.constraints = constraints
        self._dispatch = 16

cdef class ConstrainedFloatMethod(FloatMethod):
    cdef readonly tuple constraints

    cpdef deserialize(self, object data):
        return ConstrainedFloatMethod_deserialize(self, data)

    def __init__(self, constraints):
        self.constraints = constraints
        self._dispatch = 18

cdef class ConstrainedStrMethod(StrMethod):
    cdef readonly tuple constraints

    cpdef deserialize(self, object data):
        return ConstrainedStrMethod_deserialize(self, data)

    def __init__(self, constraints):
        self.constraints = constraints
        self._dispatch = 20

cdef class SubprimitiveMethod(DeserializationMethod):
    cdef readonly type cls
    cdef readonly DeserializationMethod method

    cpdef deserialize(self, object data):
        return SubprimitiveMethod_deserialize(self, data)

    def __init__(self, cls, method):
        self.cls = cls
        self.method = method
        self._dispatch = 22

cdef class TupleMethod(DeserializationMethod):
    cdef readonly tuple constraints
    cdef readonly object min_len_error
    cdef readonly object max_len_error
    cdef readonly tuple elt_methods

    cpdef deserialize(self, object data):
        return TupleMethod_deserialize(self, data)

    def __init__(self, constraints, min_len_error, max_len_error, elt_methods):
        self.constraints = constraints
        self.min_len_error = min_len_error
        self.max_len_error = max_len_error
        self.elt_methods = elt_methods
        self._dispatch = 23

cdef class OptionalMethod(DeserializationMethod):
    cdef readonly DeserializationMethod value_method
    cdef readonly object coercer

    cpdef deserialize(self, object data):
        return OptionalMethod_deserialize(self, data)

    def __init__(self, value_method, coercer):
        self.value_method = value_method
        self.coercer = coercer
        self._dispatch = 24

cdef class UnionByTypeMethod(DeserializationMethod):
    cdef readonly dict method_by_cls

    cpdef deserialize(self, object data):
        return UnionByTypeMethod_deserialize(self, data)

    def __init__(self, method_by_cls):
        self.method_by_cls = method_by_cls
        self._dispatch = 25

cdef class UnionMethod(DeserializationMethod):
    cdef readonly tuple alt_methods

    cpdef deserialize(self, object data):
        return UnionMethod_deserialize(self, data)

    def __init__(self, alt_methods):
        self.alt_methods = alt_methods
        self._dispatch = 26

cdef class ConversionMethod(DeserializationMethod):
    cdef readonly object converter
    cdef readonly DeserializationMethod method

    cpdef deserialize(self, object data):
        return ConversionMethod_deserialize(self, data)

    def __init__(self, converter, method):
        self.converter = converter
        self.method = method
        self._dispatch = 27

cdef class ConversionWithValueErrorMethod(ConversionMethod):

    cpdef deserialize(self, object data):
        return ConversionWithValueErrorMethod_deserialize(self, data)

    def __init__(self, converter, method):
        self.converter = converter
        self.method = method
        self._dispatch = 28

cdef class ConversionAlternative():
    cdef readonly object converter
    cdef readonly DeserializationMethod method
    cdef readonly bint value_error

    def __init__(self, converter, method, value_error):
        self.converter = converter
        self.method = method
        self.value_error = value_error

cdef class ConversionUnionMethod(DeserializationMethod):
    cdef readonly tuple alternatives

    cpdef deserialize(self, object data):
        return ConversionUnionMethod_deserialize(self, data)

    def __init__(self, alternatives):
        self.alternatives = alternatives
        self._dispatch = 29

cdef class DiscriminatorMethod(DeserializationMethod):
    cdef readonly str alias
    cdef readonly dict mapping
    cdef readonly str missing
    cdef readonly object error

    cpdef deserialize(self, object data):
        return DiscriminatorMethod_deserialize(self, data)

    def __init__(self, alias, mapping, missing, error):
        self.alias = alias
        self.mapping = mapping
        self.missing = missing
        self.error = error
        self._dispatch = 30

cpdef inline to_hashable(object data):
    if isinstance(data, list):
        return tuple(map(to_hashable, data))
    elif isinstance(data, dict):
        # Cython doesn't support tuple comprehension yet -> intermediate list
        return tuple([(k, to_hashable(data[k])) for k in sorted(data)])
    else:
        return data

cpdef inline format_error(object err, object data):
    return err if isinstance(err, str) else err(data)

cpdef inline validate_constraints(object data, tuple constraints, object children_errors):
    for i in range(len(constraints)):
        constraint: Constraint = constraints[i]
        if not Constraint_validate(constraint, data):
            errors: list = [format_error(constraint.error, data)]
            for j in range(i + 1, len(constraints)):
                constraint = constraints[j]
                if not Constraint_validate(constraint, data):
                    errors.append(format_error(constraint.error, data))
            raise ValidationError(errors, children_errors or {})
    if children_errors:
        raise ValidationError([], children_errors)
    return data

cpdef inline set_child_error(object errors, object key, object error):
    if errors is None:
        return {key: error}
    else:
        errors[key] = error
        return errors

cpdef inline extend_errors(object errors, object messages):
    if errors is None:
        return list(messages)
    else:
        errors.extend(messages)
        return errors

cpdef inline update_children_errors(object errors, dict children):
    if errors is None:
        return dict(children)
    else:
        errors.update(children)
        return errors

cdef inline MinimumConstraint_validate(MinimumConstraint self, long data):
        return data >= self.minimum

cdef inline MaximumConstraint_validate(MaximumConstraint self, long data):
        return data <= self.maximum

cdef inline ExclusiveMinimumConstraint_validate(ExclusiveMinimumConstraint self, long data):
        return data > self.exc_min

cdef inline ExclusiveMaximumConstraint_validate(ExclusiveMaximumConstraint self, long data):
        return data < self.exc_max

cdef inline MultipleOfConstraint_validate(MultipleOfConstraint self, long data):
        return not (data % self.mult_of)

cdef inline MinLengthConstraint_validate(MinLengthConstraint self, str data):
        return len(data) >= self.min_len

cdef inline MaxLengthConstraint_validate(MaxLengthConstraint self, str data):
        return len(data) <= self.max_len

cdef inline PatternConstraint_validate(PatternConstraint self, str data):
        return self.pattern.match(data) is not None

cdef inline MinItemsConstraint_validate(MinItemsConstraint self, list data):
        return len(data) >= self.min_items

cdef inline MaxItemsConstraint_validate(MaxItemsConstraint self, list data):
        return len(data) <= self.max_items

cdef inline UniqueItemsConstraint_validate(UniqueItemsConstraint self, list data):
        return len(set(map(to_hashable, data))) == len(data)

cdef inline MinPropertiesConstraint_validate(MinPropertiesConstraint self, dict data):
        return len(data) >= self.min_properties

cdef inline MaxPropertiesConstraint_validate(MaxPropertiesConstraint self, dict data):
        return len(data) <= self.max_properties

cdef inline RecMethod_deserialize(RecMethod self, object data):
        if self.method is None:
            self.method = self.lazy()
        return DeserializationMethod_deserialize(self.method, data)

cdef inline ValidatorMethod_deserialize(ValidatorMethod self, object data):
        return validate(
            DeserializationMethod_deserialize(self.method, data), self.validators, aliaser=self.aliaser
        )

cdef inline CoercerMethod_deserialize(CoercerMethod self, object data):
        return DeserializationMethod_deserialize(self.method, self.coercer(self.cls, data))

cdef inline TypeCheckMethod_deserialize(TypeCheckMethod self, object data):
        if isinstance(data, self.expected):
            return data
        return DeserializationMethod_deserialize(self.fallback, data)

cdef inline AnyMethod_deserialize(AnyMethod self, object data):
        if type(data) in self.constraints:
            validate_constraints(data, self.constraints[type(data)], None)
        return data

cdef inline ListCheckOnlyMethod_deserialize(ListCheckOnlyMethod self, object data):
        if not isinstance(data, list):
            raise bad_type(data, list)
        data2: list = data
        elt_errors = None
        for i, elt in enumerate(data2):
            try:
                DeserializationMethod_deserialize(self.value_method, elt)
            except ValidationError as err:
                elt_errors = set_child_error(elt_errors, i, err)
        validate_constraints(data2, self.constraints, elt_errors)
        return data2

cdef inline ListMethod_deserialize(ListMethod self, object data):
        if not isinstance(data, list):
            raise bad_type(data, list)
        data2: list = data
        elt_errors = None
        values: list = [None] * len(data2)
        for i, elt in enumerate(data2):
            try:
                values[i] = DeserializationMethod_deserialize(self.value_method, elt)
            except ValidationError as err:
                elt_errors = set_child_error(elt_errors, i, err)
        validate_constraints(data2, self.constraints, elt_errors)
        return values

cdef inline SetMethod_deserialize(SetMethod self, object data):
        if not isinstance(data, list):
            raise bad_type(data, list)
        data2: list = data
        elt_errors: dict = {}
        values: set = set()
        for i, elt in enumerate(data2):
            try:
                values.add(DeserializationMethod_deserialize(self.value_method, elt))
            except ValidationError as err:
                elt_errors = set_child_error(elt_errors, i, err)
        validate_constraints(data2, self.constraints, elt_errors)
        return values

cdef inline VariadicTupleMethod_deserialize(VariadicTupleMethod self, object data):
        return tuple(DeserializationMethod_deserialize(self.method, data))

cdef inline LiteralMethod_deserialize(LiteralMethod self, object data):
        try:
            return self.value_map[data]
        except KeyError:
            if self.coercer is not None:
                for cls in self.types:
                    try:
                        return self.value_map[self.coercer(cls, data)]
                    except IndexError:
                        pass
            raise ValidationError(format_error(self.error, data))
        except TypeError:
            raise bad_type(data, *self.types)

cdef inline MappingCheckOnly_deserialize(MappingCheckOnly self, object data):
        if not isinstance(data, dict):
            raise bad_type(data, dict)
        data2: dict = data
        item_errors = None
        for key, value in data2.items():
            try:
                DeserializationMethod_deserialize(self.key_method, key)
                DeserializationMethod_deserialize(self.value_method, value)
            except ValidationError as err:
                item_errors = set_child_error(item_errors, key, err)
        validate_constraints(data2, self.constraints, item_errors)
        return data2

cdef inline MappingMethod_deserialize(MappingMethod self, object data):
        if not isinstance(data, dict):
            raise bad_type(data, dict)
        data2: dict = data
        item_errors = None
        items: dict = {}
        for key, value in data2.items():
            try:
                items[DeserializationMethod_deserialize(self.key_method, key)] = DeserializationMethod_deserialize(self.value_method, 
                    value
                )
            except ValidationError as err:
                item_errors = set_child_error(item_errors, key, err)
        validate_constraints(data2, self.constraints, item_errors)
        return items

cdef inline SimpleObjectMethod_deserialize(SimpleObjectMethod self, object data):
        if not isinstance(data, dict):
            raise bad_type(data, dict)
        data2: dict = data
        fields_count = 0
        field_errors = None
        for i in range(len(self.fields)):
            field: Field = self.fields[i]
            if field.alias in data2:
                fields_count += 1
                try:
                    DeserializationMethod_deserialize(field.method, data2[field.alias])
                except ValidationError as err:
                    if field.required or not field.fall_back_on_default:
                        field_errors = set_child_error(field_errors, field.alias, err)
            elif field.required:
                field_errors = set_child_error(
                    field_errors, field.alias, ValidationError(self.missing)
                )
        if len(data2) != fields_count and not self.typed_dict:
            for key in data2.keys() - self.all_aliases:
                field_errors = set_child_error(
                    field_errors, key, ValidationError(self.unexpected)
                )
        if field_errors:
            raise ValidationError([], field_errors)
        return Constructor_construct(self.constructor, data2)

cdef inline ObjectMethod_deserialize(ObjectMethod self, object data):
        if not isinstance(data, dict):
            raise bad_type(data, dict)
        data2: dict = data
        values: dict = {}
        fields_count = 0
        errors = None
        try:
            validate_constraints(data, self.constraints, None)
        except ValidationError as err:
            errors = list(err.messages)
        field_errors = None
        for i in range(len(self.fields)):
            field: Field = self.fields[i]
            if field.required:
                try:
                    value: object = data2[field.alias]
                except KeyError:
                    field_errors = set_child_error(
                        field_errors, field.alias, ValidationError(self.missing)
                    )
                else:
                    fields_count += 1
                    try:
                        values[field.name] = DeserializationMethod_deserialize(field.method, value)
                    except ValidationError as err:
                        field_errors = set_child_error(field_errors, field.alias, err)
            elif field.alias in data2:
                fields_count += 1
                try:
                    values[field.name] = DeserializationMethod_deserialize(field.method, data2[field.alias])
                except ValidationError as err:
                    if not field.fall_back_on_default:
                        field_errors = set_child_error(field_errors, field.alias, err)
            elif field.required_by is not None and not field.required_by.isdisjoint(
                data2
            ):
                requiring: list = sorted(field.required_by & data2.keys())
                msg: str = self.missing + f" (required by {requiring})"
                field_errors = set_child_error(
                    field_errors, field.alias, ValidationError([msg])
                )
        if self.aggregate_fields:
            remain = data2.keys() - self.all_aliases
            for i in range(len(self.flattened_fields)):
                flattened_field: FlattenedField = self.flattened_fields[i]
                flattened: dict = {
                    alias: data2[alias]
                    for alias in flattened_field.aliases
                    if alias in data2
                }
                remain.difference_update(flattened)
                try:
                    values[flattened_field.name] = DeserializationMethod_deserialize(flattened_field.method, 
                        flattened
                    )
                except ValidationError as err:
                    if not flattened_field.fall_back_on_default:
                        errors = extend_errors(errors, err.messages)
                        field_errors = update_children_errors(
                            field_errors, err.children
                        )
            for i in range(len(self.pattern_fields)):
                pattern_field: PatternField = self.pattern_fields[i]
                matched: dict = {
                    key: data2[key]
                    for key in remain
                    if pattern_field.pattern.match(key)
                }
                remain.difference_update(matched)
                try:
                    values[pattern_field.name] = DeserializationMethod_deserialize(pattern_field.method, 
                        matched
                    )
                except ValidationError as err:
                    if not pattern_field.fall_back_on_default:
                        errors = extend_errors(errors, err.messages)
                        field_errors = update_children_errors(
                            field_errors, err.children
                        )
            if self.additional_field is not None:
                additional: dict = {key: data2[key] for key in remain}
                try:
                    values[
                        self.additional_field.name
                    ] = DeserializationMethod_deserialize(self.additional_field.method, additional)
                except ValidationError as err:
                    if not self.additional_field.fall_back_on_default:
                        errors = extend_errors(errors, err.messages)
                        field_errors = update_children_errors(
                            field_errors, err.children
                        )
            elif remain:
                if not self.additional_properties:
                    for key in remain:
                        if key != self.discriminator:
                            field_errors = set_child_error(
                                field_errors, key, ValidationError(self.unexpected)
                            )
                elif self.typed_dict:
                    for key in remain:
                        values[key] = data2[key]
        elif len(data2) != fields_count:
            if not self.additional_properties:
                for key in data2.keys() - self.all_aliases:
                    if key != self.discriminator:
                        field_errors = set_child_error(
                            field_errors, key, ValidationError(self.unexpected)
                        )
            elif self.typed_dict:
                for key in data2.keys() - self.all_aliases:
                    values[key] = data2[key]
        if self.validators:
            init = None
            if self.init_defaults:
                init = {}
                for name, default_factory in self.init_defaults:
                    if name in values:
                        init[name] = values[name]
                    elif not field_errors or name not in field_errors:
                        assert default_factory is not None
                        init[name] = default_factory()
            aliases = values.keys()
            # Don't keep validators when all dependencies are default
            validators = [
                v for v in self.validators if not v.dependencies.isdisjoint(aliases)
            ]
            if field_errors or errors:
                error = ValidationError(errors or [], field_errors or {})
                invalid_fields = self.post_init_modified
                if field_errors:
                    invalid_fields |= field_errors.keys()
                try:
                    valid_validators = [
                        v
                        for v in validators
                        if v.dependencies.isdisjoint(invalid_fields)
                    ]
                    validate(
                        ValidatorMock(self.constructor.cls, values),
                        valid_validators,
                        init,
                        aliaser=self.aliaser,
                    )
                except ValidationError as err:
                    error = merge_errors(error, err)
                raise error
            obj = Constructor_construct(self.constructor, values)
            return validate(obj, validators, init, aliaser=self.aliaser)
        elif field_errors or errors:
            raise ValidationError(errors or [], field_errors or {})
        return Constructor_construct(self.constructor, values)

cdef inline NoneMethod_deserialize(NoneMethod self, object data):
        if data is not None:
            raise bad_type(data, NoneType)
        return data

cdef inline IntMethod_deserialize(IntMethod self, object data):
        if not isinstance(data, int):
            raise bad_type(data, int)
        return data

cdef inline ConstrainedIntMethod_deserialize(ConstrainedIntMethod self, object data):
        return validate_constraints(IntMethod_deserialize(<IntMethod>self, data), self.constraints, None)

cdef inline FloatMethod_deserialize(FloatMethod self, object data):
        if isinstance(data, float):
            return data
        elif isinstance(data, int):
            return float(data)
        else:
            raise bad_type(data, float)

cdef inline ConstrainedFloatMethod_deserialize(ConstrainedFloatMethod self, object data):
        return validate_constraints(FloatMethod_deserialize(<FloatMethod>self, data), self.constraints, None)

cdef inline StrMethod_deserialize(StrMethod self, object data):
        if not isinstance(data, str):
            raise bad_type(data, str)
        return data

cdef inline ConstrainedStrMethod_deserialize(ConstrainedStrMethod self, object data):
        return validate_constraints(StrMethod_deserialize(<StrMethod>self, data), self.constraints, None)

cdef inline BoolMethod_deserialize(BoolMethod self, object data):
        if not isinstance(data, bool):
            raise bad_type(data, bool)
        return data

cdef inline SubprimitiveMethod_deserialize(SubprimitiveMethod self, object data):
        return self.cls(DeserializationMethod_deserialize(self.method, data))

cdef inline TupleMethod_deserialize(TupleMethod self, object data):
        if not isinstance(data, list):
            raise bad_type(data, list)
        data2: list = data
        if len(data2) != len(self.elt_methods):
            if len(data2) < len(self.elt_methods):
                raise ValidationError(format_error(self.min_len_error, data2))
            elif len(data2) > len(self.elt_methods):
                raise ValidationError(format_error(self.max_len_error, data2))
            else:
                raise NotImplementedError
        elt_errors: dict = {}
        elts: list = [None] * len(self.elt_methods)
        for i in range(len(self.elt_methods)):
            elt_method: DeserializationMethod = self.elt_methods[i]
            try:
                elts[i] = DeserializationMethod_deserialize(elt_method, data2[i])
            except ValidationError as err:
                elt_errors[i] = err
        validate_constraints(data2, self.constraints, elt_errors)
        return tuple(elts)

cdef inline OptionalMethod_deserialize(OptionalMethod self, object data):
        if data is None:
            return None
        try:
            return DeserializationMethod_deserialize(self.value_method, data)
        except ValidationError as err:
            if self.coercer is not None and self.coercer(NoneType, data) is None:
                return None
            else:
                raise merge_errors(err, bad_type(data, NoneType))

cdef inline UnionByTypeMethod_deserialize(UnionByTypeMethod self, object data):
        try:
            method: DeserializationMethod = self.method_by_cls[type(data)]
            return DeserializationMethod_deserialize(method, data)
        except KeyError:
            raise bad_type(data, *self.method_by_cls) from None
        except ValidationError as err:
            other_classes = (cls for cls in self.method_by_cls if cls is not type(data))
            raise merge_errors(err, bad_type(data, *other_classes))

cdef inline UnionMethod_deserialize(UnionMethod self, object data):
        error = None
        for i in range(len(self.alt_methods)):
            alt_method: DeserializationMethod = self.alt_methods[i]
            try:
                return DeserializationMethod_deserialize(alt_method, data)
            except ValidationError as err:
                error = merge_errors(error, err)
        assert error is not None
        raise error

cdef inline ConversionMethod_deserialize(ConversionMethod self, object data):
        return self.converter(DeserializationMethod_deserialize(self.method, data))

cdef inline ConversionWithValueErrorMethod_deserialize(ConversionWithValueErrorMethod self, object data):
        value = DeserializationMethod_deserialize(self.method, data)
        try:
            return self.converter(value)
        except ValueError as err:
            raise ValidationError(str(err))

cdef inline ConversionUnionMethod_deserialize(ConversionUnionMethod self, object data):
        error: Optional[ValidationError] = None
        for i in range(len(self.alternatives)):
            alternative: ConversionAlternative = self.alternatives[i]
            try:
                value = DeserializationMethod_deserialize(alternative.method, data)
            except ValidationError as err:
                error = merge_errors(error, err)
                continue
            try:
                return alternative.converter(value)
            except ValidationError as err:
                error = merge_errors(error, err)
            except ValueError as err:
                if not alternative.value_error:
                    raise
                error = merge_errors(error, ValidationError(str(err)))
        assert error is not None
        raise error

cdef inline DiscriminatorMethod_deserialize(DiscriminatorMethod self, object data):
        if not isinstance(data, dict):
            raise bad_type(data, dict)
        data2: dict = data
        if self.alias not in data2:
            raise ValidationError([], {self.alias: ValidationError(self.missing)})
        try:
            method: DeserializationMethod = self.mapping[data2[self.alias]]
        except (TypeError, KeyError):
            raise ValidationError(
                [],
                {
                    self.alias: ValidationError(
                        format_error(self.error, data2[self.alias])
                    )
                },
            )
        else:
            return DeserializationMethod_deserialize(method, data)

cdef inline NoConstructor_construct(NoConstructor self, dict fields):
        return fields

cdef inline RawConstructor_construct(RawConstructor self, dict fields):
        return self.cls(**fields)

cdef inline FieldsConstructor_construct(FieldsConstructor self, dict fields):
        obj: object = object.__new__(self.cls)
        obj_dict: dict = obj.__dict__
        obj_dict.update(fields)
        if len(fields) != self.nb_fields:
            for i in range(len(self.default_fields)):
                default_field: DefaultField = self.default_fields[i]
                if default_field.name not in fields:
                    obj_dict[default_field.name] = default_field.default_value
            for i in range(len(self.factory_fields)):
                factory_field: FactoryField = self.factory_fields[i]
                if factory_field.name not in fields:
                    obj_dict[factory_field.name] = factory_field.factory()
        return obj

cdef inline Constraint_validate(Constraint self, object data):
    cdef int _dispatch = self._dispatch
    if _dispatch == 0:
        return MinimumConstraint_validate(<MinimumConstraint>self, data)
    elif _dispatch == 1:
        return MaximumConstraint_validate(<MaximumConstraint>self, data)
    elif _dispatch == 2:
        return ExclusiveMinimumConstraint_validate(<ExclusiveMinimumConstraint>self, data)
    elif _dispatch == 3:
        return ExclusiveMaximumConstraint_validate(<ExclusiveMaximumConstraint>self, data)
    elif _dispatch == 4:
        return MultipleOfConstraint_validate(<MultipleOfConstraint>self, data)
    elif _dispatch == 5:
        return MinLengthConstraint_validate(<MinLengthConstraint>self, data)
    elif _dispatch == 6:
        return MaxLengthConstraint_validate(<MaxLengthConstraint>self, data)
    elif _dispatch == 7:
        return PatternConstraint_validate(<PatternConstraint>self, data)
    elif _dispatch == 8:
        return MinItemsConstraint_validate(<MinItemsConstraint>self, data)
    elif _dispatch == 9:
        return MaxItemsConstraint_validate(<MaxItemsConstraint>self, data)
    elif _dispatch == 10:
        return UniqueItemsConstraint_validate(<UniqueItemsConstraint>self, data)
    elif _dispatch == 11:
        return MinPropertiesConstraint_validate(<MinPropertiesConstraint>self, data)
    elif _dispatch == 12:
        return MaxPropertiesConstraint_validate(<MaxPropertiesConstraint>self, data)

cdef inline DeserializationMethod_deserialize(DeserializationMethod self, object data):
    cdef int _dispatch = self._dispatch
    if _dispatch == 0:
        return RecMethod_deserialize(<RecMethod>self, data)
    elif _dispatch == 1:
        return ValidatorMethod_deserialize(<ValidatorMethod>self, data)
    elif _dispatch == 2:
        return CoercerMethod_deserialize(<CoercerMethod>self, data)
    elif _dispatch == 3:
        return TypeCheckMethod_deserialize(<TypeCheckMethod>self, data)
    elif _dispatch == 4:
        return AnyMethod_deserialize(<AnyMethod>self, data)
    elif _dispatch == 5:
        return ListCheckOnlyMethod_deserialize(<ListCheckOnlyMethod>self, data)
    elif _dispatch == 6:
        return ListMethod_deserialize(<ListMethod>self, data)
    elif _dispatch == 7:
        return SetMethod_deserialize(<SetMethod>self, data)
    elif _dispatch == 8:
        return VariadicTupleMethod_deserialize(<VariadicTupleMethod>self, data)
    elif _dispatch == 9:
        return LiteralMethod_deserialize(<LiteralMethod>self, data)
    elif _dispatch == 10:
        return MappingCheckOnly_deserialize(<MappingCheckOnly>self, data)
    elif _dispatch == 11:
        return MappingMethod_deserialize(<MappingMethod>self, data)
    elif _dispatch == 12:
        return SimpleObjectMethod_deserialize(<SimpleObjectMethod>self, data)
    elif _dispatch == 13:
        return ObjectMethod_deserialize(<ObjectMethod>self, data)
    elif _dispatch == 14:
        return NoneMethod_deserialize(<NoneMethod>self, data)
    elif _dispatch == 15:
        return IntMethod_deserialize(<IntMethod>self, data)
    elif _dispatch == 16:
        return ConstrainedIntMethod_deserialize(<ConstrainedIntMethod>self, data)
    elif _dispatch == 17:
        return FloatMethod_deserialize(<FloatMethod>self, data)
    elif _dispatch == 18:
        return ConstrainedFloatMethod_deserialize(<ConstrainedFloatMethod>self, data)
    elif _dispatch == 19:
        return StrMethod_deserialize(<StrMethod>self, data)
    elif _dispatch == 20:
        return ConstrainedStrMethod_deserialize(<ConstrainedStrMethod>self, data)
    elif _dispatch == 21:
        return BoolMethod_deserialize(<BoolMethod>self, data)
    elif _dispatch == 22:
        return SubprimitiveMethod_deserialize(<SubprimitiveMethod>self, data)
    elif _dispatch == 23:
        return TupleMethod_deserialize(<TupleMethod>self, data)
    elif _dispatch == 24:
        return OptionalMethod_deserialize(<OptionalMethod>self, data)
    elif _dispatch == 25:
        return UnionByTypeMethod_deserialize(<UnionByTypeMethod>self, data)
    elif _dispatch == 26:
        return UnionMethod_deserialize(<UnionMethod>self, data)
    elif _dispatch == 27:
        return ConversionMethod_deserialize(<ConversionMethod>self, data)
    elif _dispatch == 28:
        return ConversionWithValueErrorMethod_deserialize(<ConversionWithValueErrorMethod>self, data)
    elif _dispatch == 29:
        return ConversionUnionMethod_deserialize(<ConversionUnionMethod>self, data)
    elif _dispatch == 30:
        return DiscriminatorMethod_deserialize(<DiscriminatorMethod>self, data)

cdef inline Constructor_construct(Constructor self, dict fields):
    cdef int _dispatch = self._dispatch
    if _dispatch == 0:
        return NoConstructor_construct(<NoConstructor>self, fields)
    elif _dispatch == 1:
        return RawConstructor_construct(<RawConstructor>self, fields)
    elif _dispatch == 2:
        return FieldsConstructor_construct(<FieldsConstructor>self, fields)




import collections.abc
from contextlib import suppress
from dataclasses import dataclass
from enum import Enum
from functools import lru_cache
from typing import (
    Any,
    Callable,
    Collection,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from apischema.aliases import Aliaser
from apischema.cache import cache
from apischema.conversions.conversions import AnyConversion, DefaultConversion
from apischema.conversions.utils import Converter
from apischema.conversions.visitor import (
    Serialization,
    SerializationVisitor,
    sub_conversion,
)
from apischema.discriminators import Discriminator, get_inherited_discriminator
from apischema.fields import support_fields_set
from apischema.metadata.keys import DISCRIMINATOR_METADATA
from apischema.objects import AliasedStr, ObjectField, object_fields
from apischema.objects.visitor import SerializationObjectVisitor
from apischema.ordering import Ordering, sort_by_order
from apischema.recursion import RecursiveConversionsVisitor
from apischema.serialization.methods import (
    AnyFallback,
    AnyMethod,
    BaseField,
    BoolMethod,
    CheckedTupleMethod,
    ClassMethod,
    ClassWithFieldsSetMethod,
    CollectionCheckOnlyMethod,
    CollectionMethod,
    ComplexField,
    ConversionMethod,
    DictMethod,
    DiscriminatedAlternative,
    DiscriminateTypedDict,
    EnumMethod,
    Fallback,
    FloatMethod,
    IdentityField,
    IdentityMethod,
    IntMethod,
    ListMethod,
    MappingCheckOnlyMethod,
    MappingMethod,
    NoFallback,
    NoneMethod,
    OptionalMethod,
    RecMethod,
    SerializationMethod,
    SerializedField,
    SimpleField,
    StrMethod,
    TupleCheckOnlyMethod,
    TupleMethod,
    TypeCheckIdentityMethod,
    TypeCheckMethod,
    TypedDictMethod,
    TypedDictWithAdditionalMethod,
    UnionAlternative,
    UnionMethod,
    ValueMethod,
    WrapperMethod,
)
from apischema.serialization.serialized_methods import get_serialized_methods
from apischema.types import AnyType, NoneType, Undefined, UndefinedType
from apischema.typing import (
    get_args,
    get_origin,
    is_new_type,
    is_type,
    is_type_var,
    is_typed_dict,
    is_union,
)
from apischema.utils import (
    CollectionOrPredicate,
    Lazy,
    as_predicate,
    deprecate_kwargs,
    get_origin_or_type,
    get_origin_or_type2,
    identity,
    is_union_of,
    opt_or,
)
from apischema.visitor import Unsupported

IDENTITY_METHOD = IdentityMethod()

METHODS = {
    identity: IDENTITY_METHOD,
    list: ListMethod(),
    dict: DictMethod(),
    str: StrMethod(),
    int: IntMethod(),
    bool: BoolMethod(),
    float: FloatMethod(),
    NoneType: NoneMethod(),
}

SerializationMethodFactory = Callable[[AnyType], SerializationMethod]

T = TypeVar("T")


def expected_class(tp: AnyType) -> type:
    origin = get_origin_or_type2(tp)
    if origin is NoneType:
        return NoneType
    elif is_typed_dict(origin):
        return collections.abc.Mapping
    elif is_type(origin):
        return origin
    elif is_new_type(origin):
        return expected_class(origin.__supertype__)
    elif is_type_var(origin) or origin is Any:
        return object
    else:
        raise TypeError(f"{tp} is not supported in union serialization")


@dataclass(frozen=True)
class PassThroughOptions:
    any: bool = False
    collections: bool = False
    enums: bool = False
    tuple: bool = False
    types: CollectionOrPredicate[AnyType] = ()

    def __post_init__(self):
        if isinstance(self.types, Collection) and not isinstance(self.types, tuple):
            object.__setattr__(self, "types", tuple(self.types))
        if self.collections and not self.tuple:
            object.__setattr__(self, "tuple", True)


@dataclass
class FieldToOrder:
    name: str
    ordering: Optional[Ordering]
    field: BaseField


CHECK_ONLY_METHODS = (
    IdentityMethod,
    TypeCheckIdentityMethod,
    CollectionCheckOnlyMethod,
    MappingCheckOnlyMethod,
)


def check_only(method: SerializationMethod) -> bool:
    """If the method transforms the data"""
    return (
        isinstance(method, CHECK_ONLY_METHODS)
        or (isinstance(method, TypeCheckMethod) and check_only(method.method))
        or (isinstance(method, OptionalMethod) and check_only(method.value_method))
        or (
            isinstance(method, UnionMethod)
            and all(check_only(alt.method) for alt in method.alternatives)
        )
    )


class SerializationMethodVisitor(
    RecursiveConversionsVisitor[Serialization, SerializationMethod],
    SerializationVisitor[SerializationMethod],
    SerializationObjectVisitor[SerializationMethod],
):
    use_cache: bool = True

    def __init__(
        self,
        additional_properties: bool,
        aliaser: Aliaser,
        check_type: bool,
        default_conversion: DefaultConversion,
        exclude_defaults: bool,
        exclude_none: bool,
        exclude_unset: bool,
        fall_back_on_any: bool,
        no_copy: bool,
        pass_through_options: PassThroughOptions,
    ):
        super().__init__(default_conversion)
        self.additional_properties = additional_properties
        self.aliaser = aliaser
        self.check_type = check_type
        self.exclude_defaults = exclude_defaults
        self.exclude_none = exclude_none
        self.exclude_unset = exclude_unset
        self.fall_back_on_any = fall_back_on_any
        self.no_copy = no_copy
        self.pass_through_options = pass_through_options
        self.pass_through_type = as_predicate(self.pass_through_options.types)

    @property
    def _factory(self) -> SerializationMethodFactory:
        return serialization_method_factory(
            self.additional_properties,
            self.aliaser,
            self.check_type,
            self._conversion,
            self.default_conversion,
            self.exclude_defaults,
            self.exclude_none,
            self.exclude_unset,
            self.fall_back_on_any,
            self.no_copy,
            self.pass_through_options,
        )

    def visit_not_recursive(self, tp: AnyType):
        return self._factory(tp) if self.use_cache else super().visit_not_recursive(tp)

    def _recursive_result(self, lazy: Lazy[SerializationMethod]) -> SerializationMethod:
        return RecMethod(lazy)

    def discriminate(self, discriminator: Discriminator, types: Sequence[Type]):
        fallback = self._any_fallback(Union[types])
        if all(map(is_typed_dict, types)):
            with suppress(Exception):
                field_names = set()
                for tp in types:
                    for field in object_fields(tp, serialization=True).values():
                        if field.alias == discriminator.alias:
                            field_names.add(field.name)
                (field_name,) = field_names
                return DiscriminateTypedDict(
                    field_name,
                    {
                        key: self.visit(tp)
                        for key, tp in discriminator.get_mapping(types).items()
                    },
                    fallback,
                )
        else:
            return UnionMethod(
                tuple(
                    DiscriminatedAlternative(
                        expected_class(tp), self.visit(tp), discriminator.alias, key
                    )
                    for key, tp in discriminator.get_mapping(types).items()
                ),
                fallback,
            )

    def annotated(self, tp: AnyType, annotations: Sequence[Any]) -> SerializationMethod:
        for annotation in reversed(annotations):
            if (
                isinstance(annotation, Mapping)
                and DISCRIMINATOR_METADATA in annotation
                and is_union(get_origin(tp))
            ):
                return self.discriminate(
                    annotation[DISCRIMINATOR_METADATA], get_args(tp)
                )
        return super().annotated(tp, annotations)

    def any(self) -> SerializationMethod:
        if self.pass_through_options.any:
            return IDENTITY_METHOD
        return AnyMethod(self._factory)

    def _any_fallback(self, tp: AnyType) -> Fallback:
        return AnyFallback(self.any()) if self.fall_back_on_any else NoFallback(tp)

    def _wrap(self, tp: AnyType, method: SerializationMethod) -> SerializationMethod:
        if not self.check_type:
            return method
        elif method is IDENTITY_METHOD:
            return TypeCheckIdentityMethod(expected_class(tp), self._any_fallback(tp))
        else:
            return TypeCheckMethod(method, expected_class(tp), self._any_fallback(tp))

    def collection(
        self, cls: Type[Collection], value_type: AnyType
    ) -> SerializationMethod:
        value_method = self.visit(value_type)

        method: SerializationMethod
        passthrough = (
            (self.no_copy and issubclass(cls, list))
            or (self.pass_through_options.tuple and issubclass(cls, tuple))
            or (
                self.pass_through_options.collections
                and not issubclass(cls, collections.abc.Set)
            )
        )
        if value_method is IDENTITY_METHOD:
            method = IDENTITY_METHOD if passthrough else METHODS[list]
        elif passthrough and check_only(value_method):
            method = CollectionCheckOnlyMethod(value_method)
        else:
            method = CollectionMethod(value_method)
        return self._wrap(cls, method)

    def enum(self, cls: Type[Enum]) -> SerializationMethod:
        method: SerializationMethod
        if self.pass_through_options.enums or issubclass(cls, (int, str)):
            method = IDENTITY_METHOD
        else:
            any_method = self.any()
            if any_method is IDENTITY_METHOD or all(
                m is IDENTITY_METHOD
                for m in map(self.visit, {elt.value.__class__ for elt in cls})
            ):
                method = ValueMethod()
            else:
                assert isinstance(any_method, AnyMethod)
                method = EnumMethod(any_method)
        return self._wrap(cls, method)

    def literal(self, values: Sequence[Any]) -> SerializationMethod:
        if self.pass_through_options.enums or all(
            isinstance(v, (int, str)) for v in values
        ):
            return IDENTITY_METHOD
        else:
            return self.any()

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> SerializationMethod:
        key_method, value_method = self.visit(key_type), self.visit(value_type)
        method: SerializationMethod
        passthrough = (
            issubclass(cls, dict) and self.no_copy
        ) or self.pass_through_options.collections
        if key_method is IDENTITY_METHOD and value_method is IDENTITY_METHOD:
            method = IDENTITY_METHOD if passthrough else METHODS[dict]
        elif passthrough and check_only(key_method) and check_only(value_method):
            method = MappingCheckOnlyMethod(key_method, value_method)
        else:
            method = MappingMethod(key_method, value_method)
        return self._wrap(cls, method)

    def object(self, tp: AnyType, fields: Sequence[ObjectField]) -> SerializationMethod:
        cls = get_origin_or_type(tp)
        fields_to_order = []
        for field in fields:
            field_alias = self.aliaser(field.alias) if not field.is_aggregate else None
            field_method = self.visit_with_conv(field.type, field.serialization)
            field_default = ... if field.required else field.get_default()
            base_field: BaseField
            if field_alias is None or field.skippable(
                self.exclude_defaults, self.exclude_none
            ):
                base_field = ComplexField(
                    field.name,
                    field_alias,  # type: ignore
                    field.required,
                    field_method,
                    field.skip.serialization_if,
                    is_union_of(field.type, UndefinedType)
                    or field_default is Undefined,
                    (is_union_of(field.type, NoneType) and self.exclude_none)
                    or field.none_as_undefined
                    or (field_default is None and self.exclude_defaults),
                    (field.skip.serialization_default or self.exclude_defaults)
                    and field_default not in (None, Undefined),
                    field_default,
                )
            elif field_method is IDENTITY_METHOD:
                base_field = IdentityField(field.name, field_alias, field.required)
            else:
                base_field = SimpleField(
                    field.name, field_alias, field.required, field_method
                )
            fields_to_order.append(FieldToOrder(field.name, field.ordering, base_field))
        for serialized, types in get_serialized_methods(tp):
            ret_type = types["return"]
            fields_to_order.append(
                FieldToOrder(
                    serialized.func.__name__,
                    serialized.ordering,
                    SerializedField(
                        self.aliaser(serialized.alias),
                        serialized.func,
                        is_union_of(ret_type, UndefinedType),
                        is_union_of(ret_type, NoneType) and self.exclude_none,
                        self.visit_with_conv(ret_type, serialized.conversion),
                    ),
                )
            )

        fields_to_order = sort_by_order(  # type: ignore
            cls, fields_to_order, lambda f: f.name, lambda f: f.ordering
        )
        base_fields = tuple(f.field for f in fields_to_order)
        method: SerializationMethod
        if is_typed_dict(cls):
            if self.additional_properties:
                method = TypedDictWithAdditionalMethod(
                    base_fields, {f.name for f in fields}, self.any()
                )
            else:
                method = TypedDictMethod(base_fields)
        elif self.exclude_unset and support_fields_set(cls):
            method = ClassWithFieldsSetMethod(base_fields)
        else:
            method = ClassMethod(base_fields)
        return self._wrap(cls, method)

    def primitive(self, cls: Type) -> SerializationMethod:
        return self._wrap(cls, IDENTITY_METHOD)

    def subprimitive(self, cls: Type, superclass: Type) -> SerializationMethod:
        if cls is AliasedStr:
            return WrapperMethod(self.aliaser)
        else:
            return super().subprimitive(cls, superclass)

    def tuple(self, types: Sequence[AnyType]) -> SerializationMethod:
        elt_methods = tuple(map(self.visit, types))
        method: SerializationMethod = TupleMethod(elt_methods)
        if self.pass_through_options.tuple:
            if all(m is IDENTITY_METHOD for m in elt_methods):
                method = IDENTITY_METHOD
            elif all(map(check_only, elt_methods)):
                method = TupleCheckOnlyMethod(elt_methods)
        if self.check_type:
            method = CheckedTupleMethod(len(types), method)
        return self._wrap(tuple, method)

    def union(self, types: Sequence[AnyType]) -> SerializationMethod:
        discriminator = get_inherited_discriminator(types)
        if discriminator is not None:
            return self.discriminate(discriminator, types)
        alternatives = []
        for tp in types:
            with suppress(Unsupported):
                method = _method = self.visit(tp)
                if isinstance(method, TypeCheckMethod):
                    _method = method.method
                elif isinstance(method, TypeCheckIdentityMethod):
                    _method = IdentityMethod()
                alt = UnionAlternative(expected_class(tp), _method)
                alternatives.append((method, alt))
        if not alternatives:
            raise Unsupported(Union[tuple(types)])  # type: ignore
        elif len(alternatives) == 1:
            return alternatives[0][0]
        elif all(meth is IDENTITY_METHOD for meth, _ in alternatives):
            return IDENTITY_METHOD
        elif len(alternatives) == 2 and NoneType in types:
            return OptionalMethod(
                next(meth for meth, alt in alternatives if alt.cls is not NoneType)
            )
        else:
            fallback = self._any_fallback(Union[types])
            return UnionMethod(tuple(alt for _, alt in alternatives), fallback)

    def unsupported(self, tp: AnyType) -> SerializationMethod:
        try:
            return super().unsupported(tp)
        except Unsupported:
            if self.fall_back_on_any and is_type(tp):
                if issubclass(tp, Mapping):
                    return self.visit(Mapping[Any, Any])
                elif issubclass(tp, Collection):
                    return self.visit(Collection[Any])
            raise

    def _visit_conversion(
        self,
        tp: AnyType,
        conversion: Serialization,
        dynamic: bool,
        next_conversion: Optional[AnyConversion],
    ) -> SerializationMethod:
        conv_method = self.visit_with_conv(
            conversion.target, sub_conversion(conversion, next_conversion)
        )
        converter = cast(Converter, conversion.converter)
        if converter is identity:
            method = conv_method
        elif conv_method is identity:
            method = METHODS.get(converter, WrapperMethod(converter))
        else:
            method = ConversionMethod(converter, conv_method)
        return self._wrap(tp, method)

    def visit_conversion(
        self,
        tp: AnyType,
        conversion: Optional[Serialization],
        dynamic: bool,
        next_conversion: Optional[AnyConversion] = None,
    ) -> SerializationMethod:
        if not dynamic and self.pass_through_type(tp):
            return self._wrap(tp, IDENTITY_METHOD)
        else:
            return super().visit_conversion(tp, conversion, dynamic, next_conversion)


@cache
def serialization_method_factory(
    additional_properties: bool,
    aliaser: Aliaser,
    check_type: bool,
    conversion: Optional[AnyConversion],
    default_conversion: DefaultConversion,
    exclude_defaults: bool,
    exclude_none: bool,
    exclude_unset: bool,
    fall_back_on_any: bool,
    no_copy: bool,
    pass_through: PassThroughOptions,
) -> SerializationMethodFactory:
    @lru_cache()
    def factory(tp: AnyType) -> SerializationMethod:
        return SerializationMethodVisitor(
            additional_properties,
            aliaser,
            check_type,
            default_conversion,
            exclude_defaults,
            exclude_none,
            exclude_unset,
            fall_back_on_any,
            no_copy,
            pass_through,
        ).visit_with_conv(tp, conversion)

    return factory


def serialization_method(
    type: AnyType,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    check_type: bool = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    exclude_defaults: bool = None,
    exclude_none: bool = None,
    exclude_unset: bool = None,
    fall_back_on_any: bool = None,
    no_copy: bool = None,
    pass_through: PassThroughOptions = None,
) -> Callable[[Any], Any]:
    from apischema import settings

    return serialization_method_factory(
        opt_or(additional_properties, settings.additional_properties),
        opt_or(aliaser, settings.aliaser),
        opt_or(check_type, settings.serialization.check_type),
        conversion,
        opt_or(default_conversion, settings.serialization.default_conversion),
        opt_or(exclude_defaults, settings.serialization.exclude_defaults),
        opt_or(exclude_none, settings.serialization.exclude_none),
        opt_or(exclude_unset, settings.serialization.exclude_unset),
        opt_or(fall_back_on_any, settings.serialization.fall_back_on_any),
        opt_or(no_copy, settings.serialization.no_copy),
        opt_or(pass_through, settings.serialization.pass_through),
    )(type).serialize


NO_OBJ = object()


@overload
def serialize(
    type: AnyType,
    obj: Any,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    check_type: bool = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    exclude_defaults: bool = None,
    exclude_none: bool = None,
    exclude_unset: bool = None,
    fall_back_on_any: bool = None,
    no_copy: bool = None,
    pass_through: PassThroughOptions = None,
) -> Any:
    ...


@overload
def serialize(
    obj: Any,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    check_type: bool = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    exclude_defaults: bool = None,
    exclude_none: bool = None,
    exclude_unset: bool = None,
    fall_back_on_any: bool = True,
    no_copy: bool = None,
    pass_through: PassThroughOptions = None,
) -> Any:
    ...


@deprecate_kwargs({"conversions": "conversion"})  # type: ignore
def serialize(
    type: AnyType = Any,
    obj: Any = NO_OBJ,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    check_type: bool = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    exclude_defaults: bool = None,
    exclude_none: bool = None,
    exclude_unset: bool = None,
    fall_back_on_any: bool = None,
    no_copy: bool = None,
    pass_through: PassThroughOptions = None,
) -> Any:
    # Handle overloaded signature without type
    if obj is NO_OBJ:
        type, obj = Any, type
        if fall_back_on_any is None:
            fall_back_on_any = True
    return serialization_method(
        type,
        additional_properties=additional_properties,
        aliaser=aliaser,
        check_type=check_type,
        conversion=conversion,
        default_conversion=default_conversion,
        exclude_defaults=exclude_defaults,
        exclude_none=exclude_none,
        exclude_unset=exclude_unset,
        fall_back_on_any=fall_back_on_any,
        no_copy=no_copy,
        pass_through=pass_through,
    )(obj)


def serialization_default(
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    default_conversion: DefaultConversion = None,
    exclude_defaults: bool = None,
    exclude_none: bool = None,
    exclude_unset: bool = None,
) -> Callable[[Any], Any]:
    from apischema import settings

    factory = serialization_method_factory(
        opt_or(additional_properties, settings.additional_properties),
        opt_or(aliaser, settings.aliaser),
        False,
        None,
        opt_or(default_conversion, settings.serialization.default_conversion),
        opt_or(exclude_defaults, settings.serialization.exclude_defaults),
        opt_or(exclude_none, settings.serialization.exclude_none),
        opt_or(exclude_unset, settings.serialization.exclude_unset),
        False,
        True,
        PassThroughOptions(any=True),
    )

    def method(obj: Any) -> Any:
        return factory(obj.__class__).serialize(obj)

    return method



from typing import Sequence, Union


class TypeCheckError(TypeError):
    def __init__(self, msg: str, loc: Sequence[Union[int, str]]):
        self.msg = msg
        self.loc = loc

    def __str__(self):
        return f"{list(self.loc)} {self.msg}"



__all__ = [
    "ID",
    "Mutation",
    "Query",
    "Subscription",
    "graphql_schema",
    "interface",
    "relay",
    "resolver",
]


try:
    from . import relay
    from .interfaces import interface
    from .resolvers import resolver
    from .schema import ID, Mutation, Query, Subscription, graphql_schema
except ImportError:
    raise



import collections.abc
import warnings
from contextlib import suppress
from dataclasses import dataclass
from typing import Any, Callable, MutableMapping, NamedTuple, Optional, TypeVar, Union

from apischema.cache import CacheAwareDict
from apischema.types import PRIMITIVE_TYPES, AnyType
from apischema.typing import (
    get_args,
    get_origin,
    is_named_tuple,
    is_type_var,
    is_typed_dict,
)
from apischema.utils import has_type_vars, merge_opts, replace_builtins


class TypeName(NamedTuple):
    json_schema: Optional[str] = None
    graphql: Optional[str] = None


NameOrFactory = Union[str, None, Callable[..., Optional[str]]]


def _apply_args(name_or_factory: NameOrFactory, *args) -> Optional[str]:
    return name_or_factory(*args) if callable(name_or_factory) else name_or_factory


_type_names: MutableMapping[AnyType, "TypeNameFactory"] = CacheAwareDict({})

T = TypeVar("T")


@dataclass(frozen=True)
class TypeNameFactory:
    json_schema: NameOrFactory
    graphql: NameOrFactory

    def __call__(self, tp: T) -> T:
        self.check_type(tp)
        _type_names[replace_builtins(tp)] = self
        return tp

    def check_type(self, tp: AnyType):
        if is_type_var(tp):
            raise TypeError("TypeVar cannot have a type_name")
        if has_type_vars(tp):
            if get_args(tp):
                raise TypeError("Generic alias cannot have a type_name")
            elif isinstance(self.json_schema, str) or isinstance(self.graphql, str):
                raise TypeError(
                    "Unspecialized generic type must used factory type_name"
                )

    def to_type_name(self, tp: AnyType, *args) -> TypeName:
        self.check_type(tp)
        return TypeName(
            _apply_args(self.json_schema, tp, *args),
            _apply_args(self.graphql, tp, *args),
        )


def type_name(
    ref: NameOrFactory = None,
    *,
    json_schema: NameOrFactory = None,
    graphql: NameOrFactory = None,
) -> TypeNameFactory:
    return TypeNameFactory(json_schema or ref, graphql or ref)


no_type_name = {*PRIMITIVE_TYPES, Any}


def default_type_name(tp: AnyType) -> Optional[TypeName]:
    if (
        hasattr(tp, "__name__")
        and not get_args(tp)
        and not has_type_vars(tp)
        and tp not in no_type_name
        and (
            not isinstance(tp, type)
            or not issubclass(tp, collections.abc.Collection)
            or is_named_tuple(tp)
            or is_typed_dict(tp)
        )
    ):
        return TypeName(tp.__name__, tp.__name__)
    else:
        return None


def get_type_name(tp: AnyType) -> TypeName:
    from apischema import settings

    tp = replace_builtins(tp)
    with suppress(KeyError, TypeError):
        return _type_names[tp].to_type_name(tp)
    origin, args = get_origin(tp), get_args(tp)
    if args and not has_type_vars(tp):
        with suppress(KeyError, TypeError):
            return _type_names[origin].to_type_name(origin, *args)
    return settings.default_type_name(tp) or TypeName()


@merge_opts
def merge_type_name(default: TypeName, override: TypeName) -> TypeName:
    return TypeName(
        override.json_schema or default.json_schema, override.graphql or default.graphql
    )


def schema_ref(ref: Optional[str]) -> Callable[[T], T]:
    warnings.warn("schema_ref is deprecated, use type_name instead", DeprecationWarning)
    return type_name(ref)



__all__ = [
    "JsonSchemaVersion",
    "definitions_schema",
    "deserialization_schema",
    "serialization_schema",
]

from .schema import definitions_schema, deserialization_schema, serialization_schema
from .versions import JsonSchemaVersion



cimport cython
from dataclasses import dataclass, field
from typing import AbstractSet, Any, Callable, Dict, Optional, Tuple, Union

from apischema.conversions.utils import Converter
from apischema.fields import FIELDS_SET_ATTR
from apischema.serialization.errors import TypeCheckError
from apischema.types import AnyType, Undefined
from apischema.utils import Lazy


cdef class SerializationMethod():
    cdef int _dispatch

cdef class IdentityMethod(SerializationMethod):

    cpdef serialize(self, object obj, object path = None):
        return IdentityMethod_serialize(self, obj, path)

    def __init__(self):
        self._dispatch = 0

cdef class ListMethod(SerializationMethod):

    cpdef serialize(self, object obj, object path = None):
        return ListMethod_serialize(self, obj, path)

    def __init__(self):
        self._dispatch = 1

cdef class DictMethod(SerializationMethod):

    cpdef serialize(self, object obj, object path = None):
        return DictMethod_serialize(self, obj, path)

    def __init__(self):
        self._dispatch = 2

cdef class StrMethod(SerializationMethod):

    cpdef serialize(self, object obj, object path = None):
        return StrMethod_serialize(self, obj, path)

    def __init__(self):
        self._dispatch = 3

cdef class IntMethod(SerializationMethod):

    cpdef serialize(self, object obj, object path = None):
        return IntMethod_serialize(self, obj, path)

    def __init__(self):
        self._dispatch = 4

cdef class BoolMethod(SerializationMethod):

    cpdef serialize(self, object obj, object path = None):
        return BoolMethod_serialize(self, obj, path)

    def __init__(self):
        self._dispatch = 5

cdef class FloatMethod(SerializationMethod):

    cpdef serialize(self, object obj, object path = None):
        return FloatMethod_serialize(self, obj, path)

    def __init__(self):
        self._dispatch = 6

cdef class NoneMethod(SerializationMethod):

    cpdef serialize(self, object obj, object path = None):
        return NoneMethod_serialize(self, obj, path)

    def __init__(self):
        self._dispatch = 7

cdef class RecMethod(SerializationMethod):
    cdef readonly object lazy
    cdef readonly object method

    cpdef serialize(self, object obj, object path = None):
        return RecMethod_serialize(self, obj, path)

    def __init__(self, lazy):
        self.lazy = lazy
        self.method = None
        self._dispatch = 8

cdef class AnyMethod(SerializationMethod):
    cdef readonly object factory

    cpdef serialize(self, object obj, object path = None):
        return AnyMethod_serialize(self, obj, path)

    def __init__(self, factory):
        self.factory = factory
        self._dispatch = 9

cdef class Fallback():
    cdef int _dispatch

cdef class NoFallback(Fallback):
    cdef readonly object tp

    cpdef fall_back(self, object obj, object path):
        return NoFallback_fall_back(self, obj, path)

    def __init__(self, tp):
        self.tp = tp
        self._dispatch = 0

cdef class AnyFallback(Fallback):
    cdef readonly SerializationMethod any_method

    cpdef fall_back(self, object obj, object path):
        return AnyFallback_fall_back(self, obj, path)

    def __init__(self, any_method):
        self.any_method = any_method
        self._dispatch = 1

cdef class TypeCheckIdentityMethod(SerializationMethod):
    cdef readonly object expected
    cdef readonly Fallback fallback

    cpdef serialize(self, object obj, object path = None):
        return TypeCheckIdentityMethod_serialize(self, obj, path)

    def __init__(self, expected, fallback):
        self.expected = expected
        self.fallback = fallback
        self._dispatch = 10

cdef class TypeCheckMethod(SerializationMethod):
    cdef readonly SerializationMethod method
    cdef readonly object expected
    cdef readonly Fallback fallback

    cpdef serialize(self, object obj, object path = None):
        return TypeCheckMethod_serialize(self, obj, path)

    def __init__(self, method, expected, fallback):
        self.method = method
        self.expected = expected
        self.fallback = fallback
        self._dispatch = 11

cdef class CollectionCheckOnlyMethod(SerializationMethod):
    cdef readonly SerializationMethod value_method

    cpdef serialize(self, object obj, object path = None):
        return CollectionCheckOnlyMethod_serialize(self, obj, path)

    def __init__(self, value_method):
        self.value_method = value_method
        self._dispatch = 12

cdef class CollectionMethod(SerializationMethod):
    cdef readonly SerializationMethod value_method

    cpdef serialize(self, object obj, object path = None):
        return CollectionMethod_serialize(self, obj, path)

    def __init__(self, value_method):
        self.value_method = value_method
        self._dispatch = 13

cdef class ValueMethod(SerializationMethod):

    cpdef serialize(self, object obj, object path = None):
        return ValueMethod_serialize(self, obj, path)

    def __init__(self):
        self._dispatch = 14

cdef class EnumMethod(SerializationMethod):
    cdef readonly AnyMethod any_method

    cpdef serialize(self, object obj, object path = None):
        return EnumMethod_serialize(self, obj, path)

    def __init__(self, any_method):
        self.any_method = any_method
        self._dispatch = 15

cdef class MappingCheckOnlyMethod(SerializationMethod):
    cdef readonly SerializationMethod key_method
    cdef readonly SerializationMethod value_method

    cpdef serialize(self, object obj, object path = None):
        return MappingCheckOnlyMethod_serialize(self, obj, path)

    def __init__(self, key_method, value_method):
        self.key_method = key_method
        self.value_method = value_method
        self._dispatch = 16

cdef class MappingMethod(SerializationMethod):
    cdef readonly SerializationMethod key_method
    cdef readonly SerializationMethod value_method

    cpdef serialize(self, object obj, object path = None):
        return MappingMethod_serialize(self, obj, path)

    def __init__(self, key_method, value_method):
        self.key_method = key_method
        self.value_method = value_method
        self._dispatch = 17

cdef class BaseField():
    cdef int _dispatch

cdef class IdentityField(BaseField):
    cdef readonly str name
    cdef readonly str alias
    cdef readonly bint required

    cpdef update_result(self, object obj, dict result, bint typed_dict, bint exclude_unset):
        return IdentityField_update_result(self, obj, result, typed_dict, exclude_unset)

    def __init__(self, name, alias, required):
        self.name = name
        self.alias = alias
        self.required = required
        self._dispatch = 0

cdef class SimpleField(IdentityField):
    cdef readonly SerializationMethod method

    cpdef update_result(self, object obj, dict result, bint typed_dict, bint exclude_unset):
        return SimpleField_update_result(self, obj, result, typed_dict, exclude_unset)

    def __init__(self, name, alias, required, method):
        self.name = name
        self.alias = alias
        self.required = required
        self.method = method
        self._dispatch = 1

cdef class ComplexField(SimpleField):
    cdef readonly object skip_if
    cdef readonly bint undefined
    cdef readonly bint skip_none
    cdef readonly bint skip_default
    cdef readonly object default_value
    cdef readonly bint skippable

    cpdef update_result(self, object obj, dict result, bint typed_dict, bint exclude_unset):
        return ComplexField_update_result(self, obj, result, typed_dict, exclude_unset)

    def __init__(self, name, alias, required, method, skip_if, undefined, skip_none, skip_default, default_value):
        self.name = name
        self.alias = alias
        self.required = required
        self.method = method
        self.skip_if = skip_if
        self.undefined = undefined
        self.skip_none = skip_none
        self.skip_default = skip_default
        self.default_value = default_value
        self.skippable = (
            self.skip_if or self.undefined or self.skip_none or self.skip_default
        )
        self._dispatch = 2

cdef class SerializedField(BaseField):
    cdef readonly str alias
    cdef readonly object func
    cdef readonly bint undefined
    cdef readonly bint skip_none
    cdef readonly SerializationMethod method

    cpdef update_result(self, object obj, dict result, bint typed_dict, bint exclude_unset):
        return SerializedField_update_result(self, obj, result, typed_dict, exclude_unset)

    def __init__(self, alias, func, undefined, skip_none, method):
        self.alias = alias
        self.func = func
        self.undefined = undefined
        self.skip_none = skip_none
        self.method = method
        self._dispatch = 3

cdef class ObjectMethod(SerializationMethod):
    cdef readonly tuple fields

    def __init__(self, fields):
        self.fields = fields
        self._dispatch = 18

cdef class ClassMethod(ObjectMethod):

    cpdef serialize(self, object obj, object path = None):
        return ClassMethod_serialize(self, obj, path)

    def __init__(self, fields):
        self.fields = fields
        self._dispatch = 19

cdef class ClassWithFieldsSetMethod(ObjectMethod):

    cpdef serialize(self, object obj, object path = None):
        return ClassWithFieldsSetMethod_serialize(self, obj, path)

    def __init__(self, fields):
        self.fields = fields
        self._dispatch = 20

cdef class TypedDictMethod(ObjectMethod):

    cpdef serialize(self, object obj, object path = None):
        return TypedDictMethod_serialize(self, obj, path)

    def __init__(self, fields):
        self.fields = fields
        self._dispatch = 21

cdef class TypedDictWithAdditionalMethod(TypedDictMethod):
    cdef readonly set field_names
    cdef readonly SerializationMethod any_method

    cpdef serialize(self, object obj, object path = None):
        return TypedDictWithAdditionalMethod_serialize(self, obj, path)

    def __init__(self, fields, field_names, any_method):
        self.fields = fields
        self.field_names = field_names
        self.any_method = any_method
        self._dispatch = 22

cdef class TupleCheckOnlyMethod(SerializationMethod):
    cdef readonly tuple elt_methods

    cpdef serialize(self, object obj, object path = None):
        return TupleCheckOnlyMethod_serialize(self, obj, path)

    def __init__(self, elt_methods):
        self.elt_methods = elt_methods
        self._dispatch = 23

cdef class TupleMethod(SerializationMethod):
    cdef readonly tuple elt_methods

    cpdef serialize(self, object obj, object path = None):
        return TupleMethod_serialize(self, obj, path)

    def __init__(self, elt_methods):
        self.elt_methods = elt_methods
        self._dispatch = 24

cdef class CheckedTupleMethod(SerializationMethod):
    cdef readonly long nb_elts
    cdef readonly SerializationMethod method

    cpdef serialize(self, object obj, object path = None):
        return CheckedTupleMethod_serialize(self, obj, path)

    def __init__(self, nb_elts, method):
        self.nb_elts = nb_elts
        self.method = method
        self._dispatch = 25

cdef class OptionalMethod(SerializationMethod):
    cdef readonly SerializationMethod value_method

    cpdef serialize(self, object obj, object path = None):
        return OptionalMethod_serialize(self, obj, path)

    def __init__(self, value_method):
        self.value_method = value_method
        self._dispatch = 26

cdef class UnionAlternative(SerializationMethod):
    cdef readonly object cls
    cdef readonly SerializationMethod method

    cpdef serialize(self, object obj, object path = None):
        return UnionAlternative_serialize(self, obj, path)

    def __init__(self, cls, method):
        self.cls = cls
        self.method = method
        self._dispatch = 27

cdef class DiscriminatedAlternative(UnionAlternative):
    cdef readonly str alias
    cdef readonly str key

    cpdef serialize(self, object obj, object path = None):
        return DiscriminatedAlternative_serialize(self, obj, path)

    def __init__(self, cls, method, alias, key):
        self.cls = cls
        self.method = method
        self.alias = alias
        self.key = key
        self._dispatch = 28

cdef class UnionMethod(SerializationMethod):
    cdef readonly tuple alternatives
    cdef readonly Fallback fallback

    cpdef serialize(self, object obj, object path = None):
        return UnionMethod_serialize(self, obj, path)

    def __init__(self, alternatives, fallback):
        self.alternatives = alternatives
        self.fallback = fallback
        self._dispatch = 29

cdef class WrapperMethod(SerializationMethod):
    cdef readonly object wrapped

    cpdef serialize(self, object obj, object path = None):
        return WrapperMethod_serialize(self, obj, path)

    def __init__(self, wrapped):
        self.wrapped = wrapped
        self._dispatch = 30

cdef class ConversionMethod(SerializationMethod):
    cdef readonly object converter
    cdef readonly SerializationMethod method

    cpdef serialize(self, object obj, object path = None):
        return ConversionMethod_serialize(self, obj, path)

    def __init__(self, converter, method):
        self.converter = converter
        self.method = method
        self._dispatch = 31

cdef class DiscriminateTypedDict(SerializationMethod):
    cdef readonly str field_name
    cdef readonly dict mapping
    cdef readonly Fallback fallback

    cpdef serialize(self, object obj, object path = None):
        return DiscriminateTypedDict_serialize(self, obj, path)

    def __init__(self, field_name, mapping, fallback):
        self.field_name = field_name
        self.mapping = mapping
        self.fallback = fallback
        self._dispatch = 32

cpdef inline serialize_field(IdentityField field, object obj, bint typed_dict, bint exclude_unset):
    if typed_dict:
        return field.required or field.name in obj
    else:
        return not exclude_unset or field.name in getattr(obj, FIELDS_SET_ATTR)

cpdef inline get_field_value(IdentityField field, object obj, bint typed_dict):
    return obj[field.name] if typed_dict else getattr(obj, field.name)

cdef inline IdentityMethod_serialize(IdentityMethod self, object obj, object path = None):
        return obj

cdef inline ListMethod_serialize(ListMethod self, object obj, object path = None):
        return list(obj)

cdef inline DictMethod_serialize(DictMethod self, object obj, object path = None):
        return dict(obj)

cdef inline StrMethod_serialize(StrMethod self, object obj, object path = None):
        return str(obj)

cdef inline IntMethod_serialize(IntMethod self, object obj, object path = None):
        return int(obj)

cdef inline BoolMethod_serialize(BoolMethod self, object obj, object path = None):
        return bool(obj)

cdef inline FloatMethod_serialize(FloatMethod self, object obj, object path = None):
        return float(obj)

cdef inline NoneMethod_serialize(NoneMethod self, object obj, object path = None):
        return None

cdef inline RecMethod_serialize(RecMethod self, object obj, object path = None):
        if self.method is None:
            self.method = self.lazy()
        return SerializationMethod_serialize(self.method, obj)

cdef inline AnyMethod_serialize(AnyMethod self, object obj, object path = None):
        method = self.factory(obj.__class__)  # tmp  variable for substitution
        return SerializationMethod_serialize(method, obj, path)

cdef inline TypeCheckIdentityMethod_serialize(TypeCheckIdentityMethod self, object obj, object path = None):
        return (
            obj
            if isinstance(obj, self.expected)
            else Fallback_fall_back(self.fallback, obj, path)
        )

cdef inline TypeCheckMethod_serialize(TypeCheckMethod self, object obj, object path = None):
        if isinstance(obj, self.expected):
            try:
                return SerializationMethod_serialize(self.method, obj)
            except TypeCheckError as err:
                if path is None:
                    raise
                raise TypeCheckError(err.msg, [path, *err.loc])
        else:
            return Fallback_fall_back(self.fallback, obj, path)

cdef inline CollectionCheckOnlyMethod_serialize(CollectionCheckOnlyMethod self, object obj, object path = None):
        for i, elt in enumerate(obj):
            SerializationMethod_serialize(self.value_method, elt, i)
        return obj

cdef inline CollectionMethod_serialize(CollectionMethod self, object obj, object path = None):
        return [SerializationMethod_serialize(self.value_method, elt, i) for i, elt in enumerate(obj)]

cdef inline ValueMethod_serialize(ValueMethod self, object obj, object path = None):
        return obj.value

cdef inline EnumMethod_serialize(EnumMethod self, object obj, object path = None):
        return SerializationMethod_serialize(self.any_method, obj.value)

cdef inline MappingCheckOnlyMethod_serialize(MappingCheckOnlyMethod self, object obj, object path = None):
        for key, value in obj.items():
            SerializationMethod_serialize(self.key_method, key, key)
            SerializationMethod_serialize(self.value_method, value, key)
        return obj

cdef inline MappingMethod_serialize(MappingMethod self, object obj, object path = None):
        return {
            SerializationMethod_serialize(self.key_method, key, key): SerializationMethod_serialize(self.value_method, value, key)
            for key, value in obj.items()
        }

cdef inline ClassMethod_serialize(ClassMethod self, object obj, object path = None):
        result: dict = {}
        for i in range(len(self.fields)):
            field: BaseField = self.fields[i]
            BaseField_update_result(field, obj, result, False, False)
        return result

cdef inline ClassWithFieldsSetMethod_serialize(ClassWithFieldsSetMethod self, object obj, object path = None):
        result: dict = {}
        for i in range(len(self.fields)):
            field: BaseField = self.fields[i]
            BaseField_update_result(field, obj, result, False, True)
        return result

cdef inline TypedDictMethod_serialize(TypedDictMethod self, object obj, object path = None):
        result: dict = {}
        for i in range(len(self.fields)):
            field: BaseField = self.fields[i]
            BaseField_update_result(field, obj, result, True, False)
        return result

cdef inline TypedDictWithAdditionalMethod_serialize(TypedDictWithAdditionalMethod self, object obj, object path = None):
        result: dict = TypedDictMethod_serialize(<TypedDictMethod>self, obj)
        for key, value in obj.items():
            if key not in self.field_names and isinstance(key, str):
                result[str(key)] = SerializationMethod_serialize(self.any_method, value, key)
        return result

cdef inline TupleCheckOnlyMethod_serialize(TupleCheckOnlyMethod self, tuple obj, object path = None):
        for i in range(len(self.elt_methods)):
            method: SerializationMethod = self.elt_methods[i]
            SerializationMethod_serialize(method, obj[i], i)
        return obj

cdef inline TupleMethod_serialize(TupleMethod self, tuple obj, object path = None):
        elts: list = [None] * len(self.elt_methods)
        for i in range(len(self.elt_methods)):
            method: SerializationMethod = self.elt_methods[i]
            elts[i] = SerializationMethod_serialize(method, obj[i], i)
        return elts

cdef inline CheckedTupleMethod_serialize(CheckedTupleMethod self, tuple obj, object path = None):
        if not len(obj) == self.nb_elts:
            raise TypeError(f"Expected {self.nb_elts}-tuple, found {len(obj)}-tuple")
        return SerializationMethod_serialize(self.method, obj)

cdef inline OptionalMethod_serialize(OptionalMethod self, object obj, object path = None):
        return SerializationMethod_serialize(self.value_method, obj, path) if obj is not None else None

cdef inline UnionAlternative_serialize(UnionAlternative self, object obj, object path = None):
        return SerializationMethod_serialize(self.method, obj, path)

cdef inline DiscriminatedAlternative_serialize(DiscriminatedAlternative self, object obj, object path = None):
        res = UnionAlternative_serialize(<UnionAlternative>self, obj, path)
        if isinstance(res, dict) and self.alias not in res:
            res[self.alias] = self.key
        return res

cdef inline UnionMethod_serialize(UnionMethod self, object obj, object path = None):
        for i in range(len(self.alternatives)):
            alternative: UnionAlternative = self.alternatives[i]
            if isinstance(obj, alternative.cls):
                try:
                    return SerializationMethod_serialize(alternative, obj, path)
                except Exception:
                    pass
        return Fallback_fall_back(self.fallback, obj, path)

cdef inline WrapperMethod_serialize(WrapperMethod self, object obj, object path = None):
        return self.wrapped(obj)

cdef inline ConversionMethod_serialize(ConversionMethod self, object obj, object path = None):
        return SerializationMethod_serialize(self.method, self.converter(obj))

cdef inline DiscriminateTypedDict_serialize(DiscriminateTypedDict self, object obj, object path = None):
        try:
            method: SerializationMethod = self.mapping[obj[self.field_name]]
        except Exception:
            return Fallback_fall_back(self.fallback, obj, path)
        return SerializationMethod_serialize(method, obj, path)

cdef inline NoFallback_fall_back(NoFallback self, object obj, object path):
        raise TypeCheckError(
            f"Expected {self.tp}, found {obj.__class__}",
            [path] if path is not None else [],
        )

cdef inline AnyFallback_fall_back(AnyFallback self, object obj, object key):
        return SerializationMethod_serialize(self.any_method, obj, key)

cdef inline IdentityField_update_result(IdentityField self, object obj, dict result, bint typed_dict, bint exclude_unset):
        if serialize_field(self, obj, typed_dict, exclude_unset):
            result[self.alias] = get_field_value(self, obj, typed_dict)

cdef inline SimpleField_update_result(SimpleField self, object obj, dict result, bint typed_dict, bint exclude_unset):
        if serialize_field(self, obj, typed_dict, exclude_unset):
            result[self.alias] = SerializationMethod_serialize(self.method, 
                get_field_value(self, obj, typed_dict), self.alias
            )

cdef inline ComplexField_update_result(ComplexField self, object obj, dict result, bint typed_dict, bint exclude_unset):
        if serialize_field(self, obj, typed_dict, exclude_unset):
            value: object = get_field_value(self, obj, typed_dict)
            if not self.skippable or not (
                (self.skip_if is not None and self.skip_if(value))
                or (self.undefined and value is Undefined)
                or (self.skip_none and value is None)
                or (self.skip_default and value == self.default_value)
            ):
                if self.alias is not None:
                    result[self.alias] = SerializationMethod_serialize(self.method, value, self.alias)
                else:
                    result.update(SerializationMethod_serialize(self.method, value, self.alias))

cdef inline SerializedField_update_result(SerializedField self, object obj, dict result, bint typed_dict, bint exclude_unset):
        value = self.func(obj)
        if not (self.undefined and value is Undefined) and not (
            self.skip_none and value is None
        ):
            result[self.alias] = SerializationMethod_serialize(self.method, value, self.alias)

cdef inline SerializationMethod_serialize(SerializationMethod self, object obj, object path = None):
    cdef int _dispatch = self._dispatch
    if _dispatch == 0:
        return IdentityMethod_serialize(<IdentityMethod>self, obj, path)
    elif _dispatch == 1:
        return ListMethod_serialize(<ListMethod>self, obj, path)
    elif _dispatch == 2:
        return DictMethod_serialize(<DictMethod>self, obj, path)
    elif _dispatch == 3:
        return StrMethod_serialize(<StrMethod>self, obj, path)
    elif _dispatch == 4:
        return IntMethod_serialize(<IntMethod>self, obj, path)
    elif _dispatch == 5:
        return BoolMethod_serialize(<BoolMethod>self, obj, path)
    elif _dispatch == 6:
        return FloatMethod_serialize(<FloatMethod>self, obj, path)
    elif _dispatch == 7:
        return NoneMethod_serialize(<NoneMethod>self, obj, path)
    elif _dispatch == 8:
        return RecMethod_serialize(<RecMethod>self, obj, path)
    elif _dispatch == 9:
        return AnyMethod_serialize(<AnyMethod>self, obj, path)
    elif _dispatch == 10:
        return TypeCheckIdentityMethod_serialize(<TypeCheckIdentityMethod>self, obj, path)
    elif _dispatch == 11:
        return TypeCheckMethod_serialize(<TypeCheckMethod>self, obj, path)
    elif _dispatch == 12:
        return CollectionCheckOnlyMethod_serialize(<CollectionCheckOnlyMethod>self, obj, path)
    elif _dispatch == 13:
        return CollectionMethod_serialize(<CollectionMethod>self, obj, path)
    elif _dispatch == 14:
        return ValueMethod_serialize(<ValueMethod>self, obj, path)
    elif _dispatch == 15:
        return EnumMethod_serialize(<EnumMethod>self, obj, path)
    elif _dispatch == 16:
        return MappingCheckOnlyMethod_serialize(<MappingCheckOnlyMethod>self, obj, path)
    elif _dispatch == 17:
        return MappingMethod_serialize(<MappingMethod>self, obj, path)
    elif _dispatch == 19:
        return ClassMethod_serialize(<ClassMethod>self, obj, path)
    elif _dispatch == 20:
        return ClassWithFieldsSetMethod_serialize(<ClassWithFieldsSetMethod>self, obj, path)
    elif _dispatch == 21:
        return TypedDictMethod_serialize(<TypedDictMethod>self, obj, path)
    elif _dispatch == 22:
        return TypedDictWithAdditionalMethod_serialize(<TypedDictWithAdditionalMethod>self, obj, path)
    elif _dispatch == 23:
        return TupleCheckOnlyMethod_serialize(<TupleCheckOnlyMethod>self, obj, path)
    elif _dispatch == 24:
        return TupleMethod_serialize(<TupleMethod>self, obj, path)
    elif _dispatch == 25:
        return CheckedTupleMethod_serialize(<CheckedTupleMethod>self, obj, path)
    elif _dispatch == 26:
        return OptionalMethod_serialize(<OptionalMethod>self, obj, path)
    elif _dispatch == 27:
        return UnionAlternative_serialize(<UnionAlternative>self, obj, path)
    elif _dispatch == 28:
        return DiscriminatedAlternative_serialize(<DiscriminatedAlternative>self, obj, path)
    elif _dispatch == 29:
        return UnionMethod_serialize(<UnionMethod>self, obj, path)
    elif _dispatch == 30:
        return WrapperMethod_serialize(<WrapperMethod>self, obj, path)
    elif _dispatch == 31:
        return ConversionMethod_serialize(<ConversionMethod>self, obj, path)
    elif _dispatch == 32:
        return DiscriminateTypedDict_serialize(<DiscriminateTypedDict>self, obj, path)

cdef inline Fallback_fall_back(Fallback self, object obj, object path):
    cdef int _dispatch = self._dispatch
    if _dispatch == 0:
        return NoFallback_fall_back(<NoFallback>self, obj, path)
    elif _dispatch == 1:
        return AnyFallback_fall_back(<AnyFallback>self, obj, path)

cdef inline BaseField_update_result(BaseField self, object obj, dict result, bint typed_dict, bint exclude_unset):
    cdef int _dispatch = self._dispatch
    if _dispatch == 0:
        return IdentityField_update_result(<IdentityField>self, obj, result, typed_dict, exclude_unset)
    elif _dispatch == 1:
        return SimpleField_update_result(<SimpleField>self, obj, result, typed_dict, exclude_unset)
    elif _dispatch == 2:
        return ComplexField_update_result(<ComplexField>self, obj, result, typed_dict, exclude_unset)
    elif _dispatch == 3:
        return SerializedField_update_result(<SerializedField>self, obj, result, typed_dict, exclude_unset)




from dataclasses import dataclass, field
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Union,
)

from apischema.aliases import Aliaser
from apischema.conversions.utils import Converter
from apischema.deserialization.coercion import Coercer
from apischema.json_schema.types import bad_type
from apischema.types import AnyType, NoneType
from apischema.utils import Lazy
from apischema.validation.errors import (
    ErrorKey,
    ErrorMsg,
    ValidationError,
    merge_errors,
)
from apischema.validation.mock import ValidatorMock
from apischema.validation.validators import Validator, validate


@dataclass
class Constraint:
    error: Union[str, Callable[[Any], str]]

    def validate(self, data: Any) -> bool:
        raise NotImplementedError


@dataclass
class MinimumConstraint(Constraint):
    minimum: int

    def validate(self, data: int) -> bool:
        return data >= self.minimum


@dataclass
class MaximumConstraint(Constraint):
    maximum: int

    def validate(self, data: int) -> bool:
        return data <= self.maximum


@dataclass
class ExclusiveMinimumConstraint(Constraint):
    exc_min: int

    def validate(self, data: int) -> bool:
        return data > self.exc_min


@dataclass
class ExclusiveMaximumConstraint(Constraint):
    exc_max: int

    def validate(self, data: int) -> bool:
        return data < self.exc_max


@dataclass
class MultipleOfConstraint(Constraint):
    mult_of: int

    def validate(self, data: int) -> bool:
        return not (data % self.mult_of)


@dataclass
class MinLengthConstraint(Constraint):
    min_len: int

    def validate(self, data: str) -> bool:
        return len(data) >= self.min_len


@dataclass
class MaxLengthConstraint(Constraint):
    max_len: int

    def validate(self, data: str) -> bool:
        return len(data) <= self.max_len


@dataclass
class PatternConstraint(Constraint):
    pattern: Pattern

    def validate(self, data: str) -> bool:
        return self.pattern.match(data) is not None


@dataclass
class MinItemsConstraint(Constraint):
    min_items: int

    def validate(self, data: list) -> bool:
        return len(data) >= self.min_items


@dataclass
class MaxItemsConstraint(Constraint):
    max_items: int

    def validate(self, data: list) -> bool:
        return len(data) <= self.max_items


def to_hashable(data: Any) -> Any:
    if isinstance(data, list):
        return tuple(map(to_hashable, data))
    elif isinstance(data, dict):
        # Cython doesn't support tuple comprehension yet -> intermediate list
        return tuple([(k, to_hashable(data[k])) for k in sorted(data)])
    else:
        return data


@dataclass
class UniqueItemsConstraint(Constraint):
    unique: bool

    def __post_init__(self):
        assert self.unique

    def validate(self, data: list) -> bool:
        return len(set(map(to_hashable, data))) == len(data)


@dataclass
class MinPropertiesConstraint(Constraint):
    min_properties: int

    def validate(self, data: dict) -> bool:
        return len(data) >= self.min_properties


@dataclass
class MaxPropertiesConstraint(Constraint):
    max_properties: int

    def validate(self, data: dict) -> bool:
        return len(data) <= self.max_properties


def format_error(err: Union[str, Callable[[Any], str]], data: Any) -> str:
    return err if isinstance(err, str) else err(data)


def validate_constraints(
    data: Any, constraints: Tuple[Constraint, ...], children_errors: Optional[dict]
) -> Any:
    for i in range(len(constraints)):
        constraint: Constraint = constraints[i]
        if not constraint.validate(data):
            errors: list = [format_error(constraint.error, data)]
            for j in range(i + 1, len(constraints)):
                constraint = constraints[j]
                if not constraint.validate(data):
                    errors.append(format_error(constraint.error, data))
            raise ValidationError(errors, children_errors or {})
    if children_errors:
        raise ValidationError([], children_errors)
    return data


def set_child_error(
    errors: Optional[Dict[ErrorKey, ValidationError]],
    key: ErrorKey,
    error: ValidationError,
):
    if errors is None:
        return {key: error}
    else:
        errors[key] = error
        return errors


class DeserializationMethod:
    def deserialize(self, data: Any) -> Any:
        raise NotImplementedError


@dataclass
class RecMethod(DeserializationMethod):
    lazy: Lazy[DeserializationMethod]
    method: Optional[DeserializationMethod] = field(init=False)

    def __post_init__(self):
        self.method = None

    def deserialize(self, data: Any) -> Any:
        if self.method is None:
            self.method = self.lazy()
        return self.method.deserialize(data)


@dataclass
class ValidatorMethod(DeserializationMethod):
    method: DeserializationMethod
    validators: Sequence[Validator]
    aliaser: Aliaser

    def deserialize(self, data: Any) -> Any:
        return validate(
            self.method.deserialize(data), self.validators, aliaser=self.aliaser
        )


@dataclass
class CoercerMethod(DeserializationMethod):
    coercer: Coercer
    cls: type
    method: DeserializationMethod

    def deserialize(self, data: Any) -> Any:
        return self.method.deserialize(self.coercer(self.cls, data))


@dataclass
class TypeCheckMethod(DeserializationMethod):
    expected: AnyType  # `type` would require exact match (i.e. no EnumMeta)
    fallback: DeserializationMethod

    def deserialize(self, data: Any) -> Any:
        if isinstance(data, self.expected):
            return data
        return self.fallback.deserialize(data)


@dataclass
class AnyMethod(DeserializationMethod):
    constraints: Dict[type, Tuple[Constraint, ...]]

    def deserialize(self, data: Any) -> Any:
        if type(data) in self.constraints:
            validate_constraints(data, self.constraints[type(data)], None)
        return data


@dataclass
class ListCheckOnlyMethod(DeserializationMethod):
    constraints: Tuple[Constraint, ...]
    value_method: DeserializationMethod

    def deserialize(self, data: Any) -> Any:
        if not isinstance(data, list):
            raise bad_type(data, list)
        data2: list = data
        elt_errors = None
        for i, elt in enumerate(data2):
            try:
                self.value_method.deserialize(elt)
            except ValidationError as err:
                elt_errors = set_child_error(elt_errors, i, err)
        validate_constraints(data2, self.constraints, elt_errors)
        return data2


@dataclass
class ListMethod(DeserializationMethod):
    constraints: Tuple[Constraint, ...]
    value_method: DeserializationMethod

    def deserialize(self, data: Any) -> Any:
        if not isinstance(data, list):
            raise bad_type(data, list)
        data2: list = data
        elt_errors = None
        values: list = [None] * len(data2)
        for i, elt in enumerate(data2):
            try:
                values[i] = self.value_method.deserialize(elt)
            except ValidationError as err:
                elt_errors = set_child_error(elt_errors, i, err)
        validate_constraints(data2, self.constraints, elt_errors)
        return values


@dataclass
class SetMethod(DeserializationMethod):
    constraints: Tuple[Constraint, ...]
    value_method: DeserializationMethod

    def deserialize(self, data: Any) -> Any:
        if not isinstance(data, list):
            raise bad_type(data, list)
        data2: list = data
        elt_errors: dict = {}
        values: set = set()
        for i, elt in enumerate(data2):
            try:
                values.add(self.value_method.deserialize(elt))
            except ValidationError as err:
                elt_errors = set_child_error(elt_errors, i, err)
        validate_constraints(data2, self.constraints, elt_errors)
        return values


@dataclass
class VariadicTupleMethod(DeserializationMethod):
    method: DeserializationMethod

    def deserialize(self, data: Any) -> Any:
        return tuple(self.method.deserialize(data))


@dataclass
class LiteralMethod(DeserializationMethod):
    value_map: dict
    error: Union[str, Callable[[Any], str]]
    coercer: Optional[Coercer]
    types: Tuple[type, ...]

    def deserialize(self, data: Any) -> Any:
        try:
            return self.value_map[data]
        except KeyError:
            if self.coercer is not None:
                for cls in self.types:
                    try:
                        return self.value_map[self.coercer(cls, data)]
                    except IndexError:
                        pass
            raise ValidationError(format_error(self.error, data))
        except TypeError:
            raise bad_type(data, *self.types)


@dataclass
class MappingCheckOnly(DeserializationMethod):
    constraints: Tuple[Constraint, ...]
    key_method: DeserializationMethod
    value_method: DeserializationMethod

    def deserialize(self, data: Any) -> Any:
        if not isinstance(data, dict):
            raise bad_type(data, dict)
        data2: dict = data
        item_errors = None
        for key, value in data2.items():
            try:
                self.key_method.deserialize(key)
                self.value_method.deserialize(value)
            except ValidationError as err:
                item_errors = set_child_error(item_errors, key, err)
        validate_constraints(data2, self.constraints, item_errors)
        return data2


@dataclass
class MappingMethod(DeserializationMethod):
    constraints: Tuple[Constraint, ...]
    key_method: DeserializationMethod
    value_method: DeserializationMethod

    def deserialize(self, data: Any) -> Any:
        if not isinstance(data, dict):
            raise bad_type(data, dict)
        data2: dict = data
        item_errors = None
        items: dict = {}
        for key, value in data2.items():
            try:
                items[self.key_method.deserialize(key)] = self.value_method.deserialize(
                    value
                )
            except ValidationError as err:
                item_errors = set_child_error(item_errors, key, err)
        validate_constraints(data2, self.constraints, item_errors)
        return items


@dataclass
class Field:
    name: str
    alias: str
    method: DeserializationMethod
    required: bool
    required_by: Optional[AbstractSet[str]]
    fall_back_on_default: bool


@dataclass
class FlattenedField:
    name: str
    aliases: Tuple[str, ...]
    method: DeserializationMethod
    fall_back_on_default: bool


@dataclass
class PatternField:
    name: str
    pattern: Pattern
    method: DeserializationMethod
    fall_back_on_default: bool


@dataclass
class AdditionalField:
    name: str
    method: DeserializationMethod
    fall_back_on_default: bool


@dataclass
class Constructor:
    cls: Any  # cython doesn't handle type subclasses properly

    def construct(self, fields: Dict[str, Any]) -> Any:
        raise NotImplementedError


class NoConstructor(Constructor):
    def construct(self, fields: Dict[str, Any]) -> Any:
        return fields


class RawConstructor(Constructor):
    def construct(self, fields: Dict[str, Any]) -> Any:
        return self.cls(**fields)


@dataclass
class DefaultField:
    name: str
    default_value: Any  # https://github.com/cython/cython/issues/4383


@dataclass
class FactoryField:
    name: str
    factory: Callable


@dataclass
class FieldsConstructor(Constructor):
    nb_fields: int
    default_fields: Tuple[DefaultField, ...]
    factory_fields: Tuple[FactoryField, ...]

    def construct(self, fields: Dict[str, Any]) -> Any:
        obj: object = object.__new__(self.cls)
        obj_dict: dict = obj.__dict__
        obj_dict.update(fields)
        if len(fields) != self.nb_fields:
            for i in range(len(self.default_fields)):
                default_field: DefaultField = self.default_fields[i]
                if default_field.name not in fields:
                    obj_dict[default_field.name] = default_field.default_value
            for i in range(len(self.factory_fields)):
                factory_field: FactoryField = self.factory_fields[i]
                if factory_field.name not in fields:
                    obj_dict[factory_field.name] = factory_field.factory()
        return obj


@dataclass
class SimpleObjectMethod(DeserializationMethod):
    constructor: Constructor
    fields: Tuple[Field, ...]
    all_aliases: AbstractSet[str]
    typed_dict: bool
    missing: str
    unexpected: str

    def deserialize(self, data: Any) -> Any:
        if not isinstance(data, dict):
            raise bad_type(data, dict)
        data2: dict = data
        fields_count = 0
        field_errors = None
        for i in range(len(self.fields)):
            field: Field = self.fields[i]
            if field.alias in data2:
                fields_count += 1
                try:
                    field.method.deserialize(data2[field.alias])
                except ValidationError as err:
                    if field.required or not field.fall_back_on_default:
                        field_errors = set_child_error(field_errors, field.alias, err)
            elif field.required:
                field_errors = set_child_error(
                    field_errors, field.alias, ValidationError(self.missing)
                )
        if len(data2) != fields_count and not self.typed_dict:
            for key in data2.keys() - self.all_aliases:
                field_errors = set_child_error(
                    field_errors, key, ValidationError(self.unexpected)
                )
        if field_errors:
            raise ValidationError([], field_errors)
        return self.constructor.construct(data2)


def extend_errors(
    errors: Optional[List[ErrorMsg]], messages: Sequence[ErrorMsg]
) -> List[ErrorMsg]:
    if errors is None:
        return list(messages)
    else:
        errors.extend(messages)
        return errors


def update_children_errors(
    errors: Optional[Dict[ErrorKey, ValidationError]],
    children: Mapping[ErrorKey, ValidationError],
) -> Dict[ErrorKey, ValidationError]:
    if errors is None:
        return dict(children)
    else:
        errors.update(children)
        return errors


@dataclass
class ObjectMethod(DeserializationMethod):
    constructor: Constructor
    constraints: Tuple[Constraint, ...]
    fields: Tuple[Field, ...]
    flattened_fields: Tuple[FlattenedField, ...]
    pattern_fields: Tuple[PatternField, ...]
    additional_field: Optional[AdditionalField]
    all_aliases: AbstractSet[str]
    additional_properties: bool
    typed_dict: bool
    validators: Tuple[Validator, ...]
    init_defaults: Tuple[Tuple[str, Optional[Callable[[], Any]]], ...]
    post_init_modified: AbstractSet[str]
    aliaser: Aliaser
    missing: str
    unexpected: str
    discriminator: Optional[str]
    aggregate_fields: bool = field(init=False)

    def __post_init__(self):
        self.aggregate_fields = bool(
            self.flattened_fields
            or self.pattern_fields
            or self.additional_field is not None
        )

    def deserialize(self, data: Any) -> Any:
        if not isinstance(data, dict):
            raise bad_type(data, dict)
        data2: dict = data
        values: dict = {}
        fields_count = 0
        errors = None
        try:
            validate_constraints(data, self.constraints, None)
        except ValidationError as err:
            errors = list(err.messages)
        field_errors = None
        for i in range(len(self.fields)):
            field: Field = self.fields[i]
            if field.required:
                try:
                    value: object = data2[field.alias]
                except KeyError:
                    field_errors = set_child_error(
                        field_errors, field.alias, ValidationError(self.missing)
                    )
                else:
                    fields_count += 1
                    try:
                        values[field.name] = field.method.deserialize(value)
                    except ValidationError as err:
                        field_errors = set_child_error(field_errors, field.alias, err)
            elif field.alias in data2:
                fields_count += 1
                try:
                    values[field.name] = field.method.deserialize(data2[field.alias])
                except ValidationError as err:
                    if not field.fall_back_on_default:
                        field_errors = set_child_error(field_errors, field.alias, err)
            elif field.required_by is not None and not field.required_by.isdisjoint(
                data2
            ):
                requiring: list = sorted(field.required_by & data2.keys())
                msg: str = self.missing + f" (required by {requiring})"
                field_errors = set_child_error(
                    field_errors, field.alias, ValidationError([msg])
                )
        if self.aggregate_fields:
            remain = data2.keys() - self.all_aliases
            for i in range(len(self.flattened_fields)):
                flattened_field: FlattenedField = self.flattened_fields[i]
                flattened: dict = {
                    alias: data2[alias]
                    for alias in flattened_field.aliases
                    if alias in data2
                }
                remain.difference_update(flattened)
                try:
                    values[flattened_field.name] = flattened_field.method.deserialize(
                        flattened
                    )
                except ValidationError as err:
                    if not flattened_field.fall_back_on_default:
                        errors = extend_errors(errors, err.messages)
                        field_errors = update_children_errors(
                            field_errors, err.children
                        )
            for i in range(len(self.pattern_fields)):
                pattern_field: PatternField = self.pattern_fields[i]
                matched: dict = {
                    key: data2[key]
                    for key in remain
                    if pattern_field.pattern.match(key)
                }
                remain.difference_update(matched)
                try:
                    values[pattern_field.name] = pattern_field.method.deserialize(
                        matched
                    )
                except ValidationError as err:
                    if not pattern_field.fall_back_on_default:
                        errors = extend_errors(errors, err.messages)
                        field_errors = update_children_errors(
                            field_errors, err.children
                        )
            if self.additional_field is not None:
                additional: dict = {key: data2[key] for key in remain}
                try:
                    values[
                        self.additional_field.name
                    ] = self.additional_field.method.deserialize(additional)
                except ValidationError as err:
                    if not self.additional_field.fall_back_on_default:
                        errors = extend_errors(errors, err.messages)
                        field_errors = update_children_errors(
                            field_errors, err.children
                        )
            elif remain:
                if not self.additional_properties:
                    for key in remain:
                        if key != self.discriminator:
                            field_errors = set_child_error(
                                field_errors, key, ValidationError(self.unexpected)
                            )
                elif self.typed_dict:
                    for key in remain:
                        values[key] = data2[key]
        elif len(data2) != fields_count:
            if not self.additional_properties:
                for key in data2.keys() - self.all_aliases:
                    if key != self.discriminator:
                        field_errors = set_child_error(
                            field_errors, key, ValidationError(self.unexpected)
                        )
            elif self.typed_dict:
                for key in data2.keys() - self.all_aliases:
                    values[key] = data2[key]
        if self.validators:
            init = None
            if self.init_defaults:
                init = {}
                for name, default_factory in self.init_defaults:
                    if name in values:
                        init[name] = values[name]
                    elif not field_errors or name not in field_errors:
                        assert default_factory is not None
                        init[name] = default_factory()
            aliases = values.keys()
            # Don't keep validators when all dependencies are default
            validators = [
                v for v in self.validators if not v.dependencies.isdisjoint(aliases)
            ]
            if field_errors or errors:
                error = ValidationError(errors or [], field_errors or {})
                invalid_fields = self.post_init_modified
                if field_errors:
                    invalid_fields |= field_errors.keys()
                try:
                    valid_validators = [
                        v
                        for v in validators
                        if v.dependencies.isdisjoint(invalid_fields)
                    ]
                    validate(
                        ValidatorMock(self.constructor.cls, values),
                        valid_validators,
                        init,
                        aliaser=self.aliaser,
                    )
                except ValidationError as err:
                    error = merge_errors(error, err)
                raise error
            obj = self.constructor.construct(values)
            return validate(obj, validators, init, aliaser=self.aliaser)
        elif field_errors or errors:
            raise ValidationError(errors or [], field_errors or {})
        return self.constructor.construct(values)


class NoneMethod(DeserializationMethod):
    def deserialize(self, data: Any) -> Any:
        if data is not None:
            raise bad_type(data, NoneType)
        return data


class IntMethod(DeserializationMethod):
    def deserialize(self, data: Any) -> Any:
        if not isinstance(data, int):
            raise bad_type(data, int)
        return data


class FloatMethod(DeserializationMethod):
    def deserialize(self, data: Any) -> Any:
        if isinstance(data, float):
            return data
        elif isinstance(data, int):
            return float(data)
        else:
            raise bad_type(data, float)


class StrMethod(DeserializationMethod):
    def deserialize(self, data: Any) -> Any:
        if not isinstance(data, str):
            raise bad_type(data, str)
        return data


class BoolMethod(DeserializationMethod):
    def deserialize(self, data: Any) -> Any:
        if not isinstance(data, bool):
            raise bad_type(data, bool)
        return data


@dataclass
class ConstrainedIntMethod(IntMethod):
    constraints: Tuple[Constraint, ...]

    def deserialize(self, data: Any) -> Any:
        return validate_constraints(super().deserialize(data), self.constraints, None)


@dataclass
class ConstrainedFloatMethod(FloatMethod):
    constraints: Tuple[Constraint, ...]

    def deserialize(self, data: Any) -> Any:
        return validate_constraints(super().deserialize(data), self.constraints, None)


@dataclass
class ConstrainedStrMethod(StrMethod):
    constraints: Tuple[Constraint, ...]

    def deserialize(self, data: Any) -> Any:
        return validate_constraints(super().deserialize(data), self.constraints, None)


@dataclass
class SubprimitiveMethod(DeserializationMethod):
    cls: type
    method: DeserializationMethod

    def deserialize(self, data: Any) -> Any:
        return self.cls(self.method.deserialize(data))


@dataclass
class TupleMethod(DeserializationMethod):
    constraints: Tuple[Constraint, ...]
    min_len_error: Union[str, Callable[[Any], str]]
    max_len_error: Union[str, Callable[[Any], str]]
    elt_methods: Tuple[DeserializationMethod, ...]

    def deserialize(self, data: Any) -> Any:
        if not isinstance(data, list):
            raise bad_type(data, list)
        data2: list = data
        if len(data2) != len(self.elt_methods):
            if len(data2) < len(self.elt_methods):
                raise ValidationError(format_error(self.min_len_error, data2))
            elif len(data2) > len(self.elt_methods):
                raise ValidationError(format_error(self.max_len_error, data2))
            else:
                raise NotImplementedError
        elt_errors: dict = {}
        elts: list = [None] * len(self.elt_methods)
        for i in range(len(self.elt_methods)):
            elt_method: DeserializationMethod = self.elt_methods[i]
            try:
                elts[i] = elt_method.deserialize(data2[i])
            except ValidationError as err:
                elt_errors[i] = err
        validate_constraints(data2, self.constraints, elt_errors)
        return tuple(elts)


@dataclass
class OptionalMethod(DeserializationMethod):
    value_method: DeserializationMethod
    coercer: Optional[Coercer]

    def deserialize(self, data: Any) -> Any:
        if data is None:
            return None
        try:
            return self.value_method.deserialize(data)
        except ValidationError as err:
            if self.coercer is not None and self.coercer(NoneType, data) is None:
                return None
            else:
                raise merge_errors(err, bad_type(data, NoneType))


@dataclass
class UnionByTypeMethod(DeserializationMethod):
    method_by_cls: Dict[type, DeserializationMethod]

    def deserialize(self, data: Any) -> Any:
        try:
            method: DeserializationMethod = self.method_by_cls[type(data)]
            return method.deserialize(data)
        except KeyError:
            raise bad_type(data, *self.method_by_cls) from None
        except ValidationError as err:
            other_classes = (cls for cls in self.method_by_cls if cls is not type(data))
            raise merge_errors(err, bad_type(data, *other_classes))


@dataclass
class UnionMethod(DeserializationMethod):
    alt_methods: Tuple[DeserializationMethod, ...]

    def deserialize(self, data: Any) -> Any:
        error = None
        for i in range(len(self.alt_methods)):
            alt_method: DeserializationMethod = self.alt_methods[i]
            try:
                return alt_method.deserialize(data)
            except ValidationError as err:
                error = merge_errors(error, err)
        assert error is not None
        raise error


@dataclass
class ConversionMethod(DeserializationMethod):
    converter: Converter
    method: DeserializationMethod

    def deserialize(self, data: Any) -> Any:
        return self.converter(self.method.deserialize(data))


@dataclass
class ConversionWithValueErrorMethod(ConversionMethod):
    def deserialize(self, data: Any) -> Any:
        value = self.method.deserialize(data)
        try:
            return self.converter(value)
        except ValueError as err:
            raise ValidationError(str(err))


@dataclass
class ConversionAlternative:
    converter: Converter
    method: DeserializationMethod
    value_error: bool


@dataclass
class ConversionUnionMethod(DeserializationMethod):
    alternatives: Tuple[ConversionAlternative, ...]

    def deserialize(self, data: Any) -> Any:
        error: Optional[ValidationError] = None
        for i in range(len(self.alternatives)):
            alternative: ConversionAlternative = self.alternatives[i]
            try:
                value = alternative.method.deserialize(data)
            except ValidationError as err:
                error = merge_errors(error, err)
                continue
            try:
                return alternative.converter(value)
            except ValidationError as err:
                error = merge_errors(error, err)
            except ValueError as err:
                if not alternative.value_error:
                    raise
                error = merge_errors(error, ValidationError(str(err)))
        assert error is not None
        raise error


@dataclass
class DiscriminatorMethod(DeserializationMethod):
    alias: str
    mapping: Dict[str, DeserializationMethod]
    missing: str
    error: Union[str, Callable[[Any], str]]

    def deserialize(self, data: Any):
        if not isinstance(data, dict):
            raise bad_type(data, dict)
        data2: dict = data
        if self.alias not in data2:
            raise ValidationError([], {self.alias: ValidationError(self.missing)})
        try:
            method: DeserializationMethod = self.mapping[data2[self.alias]]
        except (TypeError, KeyError):
            raise ValidationError(
                [],
                {
                    self.alias: ValidationError(
                        format_error(self.error, data2[self.alias])
                    )
                },
            )
        else:
            return method.deserialize(data)



from dataclasses import dataclass, field
from typing import AbstractSet, Any, Callable, Dict, Optional, Tuple, Union

from apischema.conversions.utils import Converter
from apischema.fields import FIELDS_SET_ATTR
from apischema.serialization.errors import TypeCheckError
from apischema.types import AnyType, Undefined
from apischema.utils import Lazy


class SerializationMethod:
    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        raise NotImplementedError


class IdentityMethod(SerializationMethod):
    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return obj


class ListMethod(SerializationMethod):
    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return list(obj)


class DictMethod(SerializationMethod):
    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return dict(obj)


class StrMethod(SerializationMethod):
    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return str(obj)


class IntMethod(SerializationMethod):
    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return int(obj)


class BoolMethod(SerializationMethod):
    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return bool(obj)


class FloatMethod(SerializationMethod):
    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return float(obj)


class NoneMethod(SerializationMethod):
    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return None


@dataclass
class RecMethod(SerializationMethod):
    lazy: Lazy[SerializationMethod]
    method: Optional[SerializationMethod] = field(init=False)

    def __post_init__(self):
        self.method = None

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        if self.method is None:
            self.method = self.lazy()
        return self.method.serialize(obj)


@dataclass
class AnyMethod(SerializationMethod):
    factory: Callable[[AnyType], SerializationMethod]

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        method = self.factory(obj.__class__)  # tmp  variable for substitution
        return method.serialize(obj, path)


class Fallback:
    def fall_back(self, obj: Any, path: Union[int, str, None]) -> Any:
        raise NotImplementedError


@dataclass
class NoFallback(Fallback):
    tp: AnyType

    def fall_back(self, obj: Any, path: Union[int, str, None]) -> Any:
        raise TypeCheckError(
            f"Expected {self.tp}, found {obj.__class__}",
            [path] if path is not None else [],
        )


@dataclass
class AnyFallback(Fallback):
    any_method: SerializationMethod

    def fall_back(self, obj: Any, key: Union[int, str, None]) -> Any:
        return self.any_method.serialize(obj, key)


@dataclass
class TypeCheckIdentityMethod(SerializationMethod):
    expected: AnyType  # `type` would require exact match (i.e. no EnumMeta)
    fallback: Fallback

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return (
            obj
            if isinstance(obj, self.expected)
            else self.fallback.fall_back(obj, path)
        )


@dataclass
class TypeCheckMethod(SerializationMethod):
    method: SerializationMethod
    expected: AnyType  # `type` would require exact match (i.e. no EnumMeta)
    fallback: Fallback

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        if isinstance(obj, self.expected):
            try:
                return self.method.serialize(obj)
            except TypeCheckError as err:
                if path is None:
                    raise
                raise TypeCheckError(err.msg, [path, *err.loc])
        else:
            return self.fallback.fall_back(obj, path)


@dataclass
class CollectionCheckOnlyMethod(SerializationMethod):
    value_method: SerializationMethod

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        for i, elt in enumerate(obj):
            self.value_method.serialize(elt, i)
        return obj


@dataclass
class CollectionMethod(SerializationMethod):
    value_method: SerializationMethod

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return [self.value_method.serialize(elt, i) for i, elt in enumerate(obj)]


class ValueMethod(SerializationMethod):
    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return obj.value


@dataclass
class EnumMethod(SerializationMethod):
    any_method: AnyMethod

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return self.any_method.serialize(obj.value)


@dataclass
class MappingCheckOnlyMethod(SerializationMethod):
    key_method: SerializationMethod
    value_method: SerializationMethod

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        for key, value in obj.items():
            self.key_method.serialize(key, key)
            self.value_method.serialize(value, key)
        return obj


@dataclass
class MappingMethod(SerializationMethod):
    key_method: SerializationMethod
    value_method: SerializationMethod

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return {
            self.key_method.serialize(key, key): self.value_method.serialize(value, key)
            for key, value in obj.items()
        }


class BaseField:
    def update_result(
        self, obj: Any, result: dict, typed_dict: bool, exclude_unset: bool
    ):
        raise NotImplementedError


@dataclass
class IdentityField(BaseField):
    name: str
    alias: str
    required: bool

    def update_result(
        self, obj: Any, result: dict, typed_dict: bool, exclude_unset: bool
    ):
        if serialize_field(self, obj, typed_dict, exclude_unset):
            result[self.alias] = get_field_value(self, obj, typed_dict)


def serialize_field(
    field: IdentityField, obj: Any, typed_dict: bool, exclude_unset: bool
) -> bool:
    if typed_dict:
        return field.required or field.name in obj
    else:
        return not exclude_unset or field.name in getattr(obj, FIELDS_SET_ATTR)


def get_field_value(field: IdentityField, obj: Any, typed_dict: bool) -> object:
    return obj[field.name] if typed_dict else getattr(obj, field.name)


@dataclass
class SimpleField(IdentityField):
    method: SerializationMethod

    def update_result(
        self, obj: Any, result: dict, typed_dict: bool, exclude_unset: bool
    ):
        if serialize_field(self, obj, typed_dict, exclude_unset):
            result[self.alias] = self.method.serialize(
                get_field_value(self, obj, typed_dict), self.alias
            )


@dataclass
class ComplexField(SimpleField):
    skip_if: Optional[Callable]
    undefined: bool
    skip_none: bool
    skip_default: bool
    default_value: Any  # https://github.com/cython/cython/issues/4383
    skippable: bool = field(init=False)

    def __post_init__(self):
        self.skippable = (
            self.skip_if or self.undefined or self.skip_none or self.skip_default
        )

    def update_result(
        self, obj: Any, result: dict, typed_dict: bool, exclude_unset: bool
    ):
        if serialize_field(self, obj, typed_dict, exclude_unset):
            value: object = get_field_value(self, obj, typed_dict)
            if not self.skippable or not (
                (self.skip_if is not None and self.skip_if(value))
                or (self.undefined and value is Undefined)
                or (self.skip_none and value is None)
                or (self.skip_default and value == self.default_value)
            ):
                if self.alias is not None:
                    result[self.alias] = self.method.serialize(value, self.alias)
                else:
                    result.update(self.method.serialize(value, self.alias))


@dataclass
class SerializedField(BaseField):
    alias: str
    func: Callable[[Any], Any]
    undefined: bool
    skip_none: bool
    method: SerializationMethod

    def update_result(
        self, obj: Any, result: dict, typed_dict: bool, exclude_unset: bool
    ):
        value = self.func(obj)
        if not (self.undefined and value is Undefined) and not (
            self.skip_none and value is None
        ):
            result[self.alias] = self.method.serialize(value, self.alias)


@dataclass
class ObjectMethod(SerializationMethod):
    fields: Tuple[BaseField, ...]


@dataclass
class ClassMethod(ObjectMethod):
    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        result: dict = {}
        for i in range(len(self.fields)):
            field: BaseField = self.fields[i]
            field.update_result(obj, result, False, False)
        return result


@dataclass
class ClassWithFieldsSetMethod(ObjectMethod):
    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        result: dict = {}
        for i in range(len(self.fields)):
            field: BaseField = self.fields[i]
            field.update_result(obj, result, False, True)
        return result


@dataclass
class TypedDictMethod(ObjectMethod):
    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        result: dict = {}
        for i in range(len(self.fields)):
            field: BaseField = self.fields[i]
            field.update_result(obj, result, True, False)
        return result


@dataclass
class TypedDictWithAdditionalMethod(TypedDictMethod):
    field_names: AbstractSet[str]
    any_method: SerializationMethod

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        result: dict = super().serialize(obj)
        for key, value in obj.items():
            if key not in self.field_names and isinstance(key, str):
                result[str(key)] = self.any_method.serialize(value, key)
        return result


@dataclass
class TupleCheckOnlyMethod(SerializationMethod):
    elt_methods: Tuple[SerializationMethod, ...]

    def serialize(self, obj: tuple, path: Union[int, str, None] = None) -> Any:
        for i in range(len(self.elt_methods)):
            method: SerializationMethod = self.elt_methods[i]
            method.serialize(obj[i], i)
        return obj


@dataclass
class TupleMethod(SerializationMethod):
    elt_methods: Tuple[SerializationMethod, ...]

    def serialize(self, obj: tuple, path: Union[int, str, None] = None) -> Any:
        elts: list = [None] * len(self.elt_methods)
        for i in range(len(self.elt_methods)):
            method: SerializationMethod = self.elt_methods[i]
            elts[i] = method.serialize(obj[i], i)
        return elts


@dataclass
class CheckedTupleMethod(SerializationMethod):
    nb_elts: int
    method: SerializationMethod

    def serialize(self, obj: tuple, path: Union[int, str, None] = None) -> Any:
        if not len(obj) == self.nb_elts:
            raise TypeError(f"Expected {self.nb_elts}-tuple, found {len(obj)}-tuple")
        return self.method.serialize(obj)


# There is no need of an OptionalIdentityMethod because it would mean that all methods
# are IdentityMethod, which gives IdentityMethod.


@dataclass
class OptionalMethod(SerializationMethod):
    value_method: SerializationMethod

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return self.value_method.serialize(obj, path) if obj is not None else None


@dataclass
class UnionAlternative(SerializationMethod):
    cls: AnyType  # `type` would require exact match (i.e. no EnumMeta)
    method: SerializationMethod

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return self.method.serialize(obj, path)


@dataclass
class DiscriminatedAlternative(UnionAlternative):
    alias: str
    key: str

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        res = super().serialize(obj, path)
        if isinstance(res, dict) and self.alias not in res:
            res[self.alias] = self.key
        return res


@dataclass
class UnionMethod(SerializationMethod):
    alternatives: Tuple[UnionAlternative, ...]
    fallback: Fallback

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        for i in range(len(self.alternatives)):
            alternative: UnionAlternative = self.alternatives[i]
            if isinstance(obj, alternative.cls):
                try:
                    return alternative.serialize(obj, path)
                except Exception:
                    pass
        return self.fallback.fall_back(obj, path)


@dataclass
class WrapperMethod(SerializationMethod):
    wrapped: Callable[[Any], Any]

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return self.wrapped(obj)


@dataclass
class ConversionMethod(SerializationMethod):
    converter: Converter
    method: SerializationMethod

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        return self.method.serialize(self.converter(obj))


@dataclass
class DiscriminateTypedDict(SerializationMethod):
    field_name: str
    mapping: Dict[str, SerializationMethod]
    fallback: Fallback

    def serialize(self, obj: Any, path: Union[int, str, None] = None) -> Any:
        try:
            method: SerializationMethod = self.mapping[obj[self.field_name]]
        except Exception:
            return self.fallback.fall_back(obj, path)
        return method.serialize(obj, path)



import operator
import sys
from dataclasses import dataclass
from functools import reduce
from typing import (
    Callable,
    Iterable,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from apischema.cache import CacheAwareDict
from apischema.conversions import Conversion, deserializer, serializer
from apischema.metadata.keys import DISCRIMINATOR_METADATA
from apischema.objects import object_fields
from apischema.type_names import get_type_name
from apischema.types import AnyType, MetadataMixin
from apischema.typing import get_args, is_literal, is_typed_dict
from apischema.utils import get_origin_or_type2, identity, no_annotated

try:
    from apischema.typing import Literal
except ImportError:
    Literal = ...  # type: ignore

Cls = TypeVar("Cls", bound=type)


def get_discriminated(alias: str, tp: AnyType) -> Sequence[str]:
    cls = get_origin_or_type2(tp)
    try:
        has_field = False
        for field in object_fields(cls).values():
            if field.alias == alias:
                has_field = True
                field_type = no_annotated(field.type)
                if is_literal(field_type):
                    if sys.version_info < (3, 7):  # py36
                        literal_args = field_type.__values__
                    else:
                        literal_args = get_args(field_type)
                    return [v for v in literal_args if isinstance(v, str)]
        if (
            is_typed_dict(cls) and not has_field
        ):  # TypedDict must have a discriminator field
            return ()
        return [name for name in [get_type_name(tp).json_schema] if name is not None]
    except TypeError:
        return ()


def default_discriminator_mapping(
    alias: str, types: Sequence[AnyType]
) -> Mapping[str, AnyType]:
    mapping = {}
    for tp in types:
        discriminated = get_discriminated(alias, tp)
        if not discriminated:
            raise TypeError(f"{tp} can't be discriminated")
        for key in discriminated:
            mapping[key] = tp
    return mapping


def rec_subclasses(cls: type) -> Iterable[type]:
    for sub_cls in cls.__subclasses__():
        yield sub_cls
        yield from rec_subclasses(sub_cls)


@dataclass(frozen=True, unsafe_hash=False)
class Discriminator(MetadataMixin):
    key = DISCRIMINATOR_METADATA
    alias: str
    mapping: Union[
        Mapping[str, AnyType], Callable[[str, Sequence[AnyType]], Mapping[str, AnyType]]
    ] = default_discriminator_mapping
    override_implicit: bool = True

    def get_mapping(self, types: Sequence[AnyType]) -> Mapping[str, AnyType]:
        default_mapping = default_discriminator_mapping(self.alias, types)
        if self.mapping is default_discriminator_mapping:
            return default_mapping
        mapping = (
            self.mapping(self.alias, types) if callable(self.mapping) else self.mapping
        )
        if self.override_implicit:
            mapping_types = set(mapping.values())
            mapping = dict(mapping)
            for key, tp in default_mapping.items():
                if tp not in mapping_types:
                    mapping[key] = tp
            return mapping
        else:
            return {**default_mapping, **mapping}

    # Make it hashable to be used in Annotated
    def __hash__(self):
        return hash(id(self))

    def __call__(self, cls: Cls) -> Cls:
        _discriminators[cls] = self
        deserializer(
            lazy=lambda: Conversion(
                identity, source=Union[tuple(rec_subclasses(cls))], target=cls
            ),
            target=cls,
        )
        serializer(
            lazy=lambda: Conversion(
                identity,
                source=cls,
                target=Union[tuple(rec_subclasses(cls))],
                inherited=False,
            ),
            source=cls,
        )
        return cls


_discriminators: MutableMapping[type, Discriminator] = CacheAwareDict({})
get_discriminator = _discriminators.get


discriminator = Discriminator


def get_discriminated_parent(cls: type) -> Optional[type]:
    for base in cls.__mro__:
        if base in _discriminators:
            return base
    return None


def get_inherited_discriminator(types: Iterable[AnyType]) -> Optional[Discriminator]:
    discriminators = [
        {
            base
            for base in getattr(get_origin_or_type2(tp), "__mro__", ())
            if base in _discriminators
        }
        for tp in types
    ]
    for cls in reduce(operator.and_, discriminators):
        return _discriminators[cls]
    return None



from apischema.utils import PREFIX


def metadata_key(key: str) -> str:
    return PREFIX + key


ALIAS_METADATA = metadata_key("alias")
ALIAS_NO_OVERRIDE_METADATA = metadata_key("alias_no_override")
CONVERSION_METADATA = metadata_key("conversion")
DEFAULT_AS_SET_METADATA = metadata_key("default_as_set")
DISCRIMINATOR_METADATA = metadata_key("discriminator")
FALL_BACK_ON_DEFAULT_METADATA = metadata_key("fall_back_on_default")
FLATTEN_METADATA = metadata_key("flattened")
INIT_VAR_METADATA = metadata_key("init_var")
NONE_AS_UNDEFINED_METADATA = metadata_key("none_as_undefined")
ORDERING_METADATA = metadata_key("ordering")
POST_INIT_METADATA = metadata_key("post_init")
PROPERTIES_METADATA = metadata_key("properties")
REQUIRED_METADATA = metadata_key("required")
SCHEMA_METADATA = metadata_key("schema")
SKIP_METADATA = metadata_key("skip")
VALIDATORS_METADATA = metadata_key("validators")



import operator as op
from dataclasses import dataclass, field, fields
from math import gcd
from typing import Any, Callable, Collection, Dict, Optional, Pattern, Tuple, TypeVar

from apischema.types import Number
from apischema.utils import merge_opts

T = TypeVar("T")
U = TypeVar("U")

CONSTRAINT_METADATA_KEY = "constraint"


@dataclass
class ConstraintMetadata:
    alias: str
    cls: type
    merge: Callable[[T, T], T]

    @property
    def field(self) -> Any:
        return field(default=None, metadata={CONSTRAINT_METADATA_KEY: self})


def constraint(alias: str, cls: type, merge: Callable[[T, T], T]) -> Any:
    return field(
        default=None,
        metadata={CONSTRAINT_METADATA_KEY: ConstraintMetadata(alias, cls, merge)},
    )


def merge_mult_of(m1: Number, m2: Number) -> Number:
    if not isinstance(m1, int) and not isinstance(m2, int):
        raise TypeError("multipleOf merging is only supported with integers")
    return m1 * m2 / gcd(m1, m2)  # type: ignore


def merge_pattern(p1: Pattern, p2: Pattern) -> Pattern:
    raise TypeError("Cannot merge patterns")


min_, max_ = min, max


@dataclass(frozen=True)
class Constraints:
    # number
    min: Optional[Number] = constraint("minimum", float, max_)
    max: Optional[Number] = constraint("maximum", float, min_)
    exc_min: Optional[Number] = constraint("exclusiveMinimum", float, max_)
    exc_max: Optional[Number] = constraint("exclusiveMaximum", float, min_)
    mult_of: Optional[Number] = constraint("multipleOf", float, merge_mult_of)
    # string
    min_len: Optional[int] = constraint("minLength", str, max_)
    max_len: Optional[int] = constraint("maxLength", str, min_)
    pattern: Optional[Pattern] = constraint("pattern", str, merge_pattern)
    # array
    min_items: Optional[int] = constraint("minItems", list, max_)
    max_items: Optional[int] = constraint("maxItems", list, min_)
    unique: Optional[bool] = constraint("uniqueItems", list, op.or_)
    # object
    min_props: Optional[int] = constraint("minProperties", dict, max_)
    max_props: Optional[int] = constraint("maxProperties", dict, min_)

    @property
    def attr_and_metata(
        self,
    ) -> Collection[Tuple[str, Optional[Any], ConstraintMetadata]]:
        return [
            (f.name, getattr(self, f.name), f.metadata[CONSTRAINT_METADATA_KEY])
            for f in fields(self)
            if CONSTRAINT_METADATA_KEY in f.metadata
        ]

    def merge_into(self, base_schema: Dict[str, Any]):
        for name, attr, metadata in self.attr_and_metata:
            if attr is not None:
                alias = metadata.alias
                if alias in base_schema:
                    base_schema[alias] = metadata.merge(attr, base_schema[alias])  # type: ignore
                else:
                    base_schema[alias] = attr


@merge_opts
def merge_constraints(c1: Constraints, c2: Constraints) -> Constraints:
    constraints: Dict[str, Any] = {}
    for name, attr1, metadata in c1.attr_and_metata:
        attr2 = getattr(c2, name)
        if attr1 is None:
            constraints[name] = attr2
        elif attr2 is None:
            constraints[name] = attr1
        else:
            constraints[name] = metadata.merge(attr1, attr2)  # type: ignore
    return Constraints(**constraints)  # type: ignore



import warnings
from inspect import Parameter
from typing import Any, Callable, Optional, Sequence, Union

from apischema import cache
from apischema.aliases import Aliaser
from apischema.conversions.conversions import DefaultConversion
from apischema.conversions.converters import (
    default_deserialization,
    default_serialization,
)
from apischema.deserialization.coercion import Coercer
from apischema.deserialization.coercion import coerce as coerce_
from apischema.json_schema import JsonSchemaVersion
from apischema.objects import ObjectField
from apischema.objects.fields import default_object_fields as default_object_fields_
from apischema.schemas import Schema
from apischema.serialization import PassThroughOptions
from apischema.type_names import TypeName
from apischema.type_names import default_type_name as default_type_name_
from apischema.types import AnyType
from apischema.utils import CollectionOrPredicate, to_camel_case


class ResetCache(type):
    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        cache.reset()


class MetaSettings(ResetCache):
    @property
    def camel_case(cls) -> bool:
        raise NotImplementedError

    @camel_case.setter
    def camel_case(cls, value: bool):
        settings.aliaser = to_camel_case if value else lambda s: s

    @property
    def default_schema(cls) -> Callable[[AnyType], Optional[Schema]]:
        warnings.warn(
            "settings.default_schema is deprecated,"
            " use settings.base_schema.type instead",
            DeprecationWarning,
        )
        assert cls is settings
        return cls.base_schema.type  # type: ignore

    @default_schema.setter
    def default_schema(cls, value: Callable[[AnyType], Optional[Schema]]):
        warnings.warn(
            "settings.default_schema is deprecated,"
            " use settings.base_schema.type instead",
            DeprecationWarning,
        )
        assert cls is settings
        cls.base_schema.type = value  # type: ignore


ConstraintError = Union[str, Callable[[Any, Any], str]]


class settings(metaclass=MetaSettings):
    additional_properties: bool = False
    aliaser: Aliaser = lambda s: s
    default_object_fields: Callable[
        [type], Optional[Sequence[ObjectField]]
    ] = default_object_fields_
    default_type_name: Callable[[AnyType], Optional[TypeName]] = default_type_name_
    json_schema_version: JsonSchemaVersion = JsonSchemaVersion.DRAFT_2020_12

    class base_schema:
        field: Callable[[AnyType, str, str], Optional[Schema]] = lambda *_: None
        method: Callable[[AnyType, Callable, str], Optional[Schema]] = lambda *_: None
        parameter: Callable[
            [Callable, Parameter, str], Optional[Schema]
        ] = lambda *_: None
        type: Callable[[AnyType], Optional[Schema]] = lambda *_: None

    class errors:
        minimum: ConstraintError = "less than {} (minimum)"
        maximum: ConstraintError = "greater than {} (maximum)"
        exclusive_minimum: ConstraintError = (
            "less than or equal to {} (exclusiveMinimum)"
        )
        exclusive_maximum: ConstraintError = (
            "greater than or equal to {} (exclusiveMinimum)"
        )
        multiple_of: ConstraintError = "not a multiple of {} (multipleOf)"

        min_length: ConstraintError = "string length lower than {} (minLength)"
        max_length: ConstraintError = "string length greater than {} (maxLength)"
        pattern: ConstraintError = "not matching pattern {} (pattern)"

        min_items: ConstraintError = "item count lower than {} (minItems)"
        max_items: ConstraintError = "item count greater than {} (maxItems)"
        unique_items: ConstraintError = "duplicate items (uniqueItems)"

        min_properties: ConstraintError = "property count lower than {} (minProperties)"
        max_properties: ConstraintError = (
            "property count greater than {} (maxProperties)"
        )

        one_of: ConstraintError = "not one of {} (oneOf)"

        unexpected_property: str = "unexpected property"
        missing_property: str = "missing property"

    class deserialization(metaclass=ResetCache):
        coerce: bool = False
        coercer: Coercer = coerce_
        default_conversion: DefaultConversion = default_deserialization
        fall_back_on_default: bool = False
        no_copy: bool = True
        override_dataclass_constructors = False
        pass_through: CollectionOrPredicate[type] = ()

    class serialization(metaclass=ResetCache):
        check_type: bool = False
        fall_back_on_any: bool = False
        default_conversion: DefaultConversion = default_serialization
        exclude_defaults: bool = False
        exclude_none: bool = False
        exclude_unset: bool = True
        no_copy: bool = True
        pass_through: PassThroughOptions = PassThroughOptions()
